# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""
import sys

from datetime import datetime

from MAPI import (
    PT_SYSTIME, MODRECIP_ADD, PT_LONG, PT_UNICODE, BMR_EQZ
)

from MAPI.Tags import (
    PR_MESSAGE_RECIPIENTS, PR_RESPONSE_REQUESTED, PR_ENTRYID,
    PR_DISPLAY_NAME_W, PR_ADDRTYPE_W, PR_EMAIL_ADDRESS_W, PR_RECIPIENT_TYPE,
    PR_RECIPIENT_FLAGS, respOrganized, recipOrganizer
)

from MAPI.Struct import (SPropValue, SOrRestriction, SBitMaskRestriction,
                         SExistRestriction, SNotRestriction)

from .attendee import Attendee
from .errors import NotFoundError, ArgumentError
from .recurrence import Recurrence, Occurrence
from .restriction import Restriction

from .compat import (
    benc as _benc, bdec as _bdec,
)
from .defs import (
    ASF_CANCELED, NR_COLOR, COLOR_NR, FB_STATUS,
    STATUS_FB, ASF_MEETING, RESPONSE_STATUS, STATUS_RESPONSE,
)
from .pidlid import (
    PidLidReminderSet, PidLidReminderDelta, PidLidAppointmentSubType,
    PidLidBusyStatus, PidLidGlobalObjectId, PidLidRecurring,
    PidLidTimeZoneStruct, PidLidTimeZoneDescription, PidLidLocation,
    PidLidAppointmentStateFlags, PidLidAppointmentColor, PidLidResponseStatus,
    PidLidAppointmentStartWhole, PidLidAppointmentEndWhole, PidLidAppointmentReplyName,
    PidLidAppointmentReplyTime, PidLidAppointmentCounterProposal,
    PidLidAppointmentProposedStartWhole, PidLidAppointmentProposedEndWhole
)

try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__+'.utils']

from . import timezone as _timezone


class Appointment(object):
    """Appointment mixin class

    Appointment-specific functionality, mixed into the :class:`Item`
    class.
    """

    @property
    def all_day(self):
        """Appointment is all-day."""
        proptag = self.store._pidlid_proptag(PidLidAppointmentSubType)
        return self._get_fast(proptag)

    @all_day.setter
    def all_day(self, value):
        self[PidLidAppointmentSubType] = value

    @property
    def busy_status(self): # TODO setter
        """Appointment busy status (free, tentative, busy, out_of_office,
        working_elsewhere or unknown)"""
        try:
            return {
                0: 'free',
                1: 'tentative',
                2: 'busy',
                3: 'out_of_office',
                4: 'working_elsewhere'
            }[self[PidLidBusyStatus]]
        except NotFoundError:
            return 'unknown'

    @property
    def show_as(self): # TODO deprecate?
        return self.busy_status

    @property
    def start(self):
        """Appointment start."""
        return self._get_fast(self.store._pidlid_proptag(PidLidAppointmentStartWhole))

    @start.setter
    def start(self, val): # TODO update/invalidate cache
        # TODO check if exists?
        self.create_prop('common:34070', val, PT_SYSTIME) # props identical
        self.create_prop('appointment:33293', val, PT_SYSTIME)
        if self.recurring:
            self.recurrence._update_offsets()

    @property
    def end(self):
        """Appointment end."""
        return self._get_fast(self.store._pidlid_proptag(PidLidAppointmentEndWhole))

    @end.setter
    def end(self, val): # TODO update/invalidate cache
        # TODO check if exists?
        self.create_prop('common:34071', val, PT_SYSTIME) # props identical
        self.create_prop('appointment:33294', val, PT_SYSTIME)
        if self.recurring:
            self.recurrence._update_offsets()

    @property
    def location(self):
        """Appointment location."""
        return self.get(PidLidLocation)

    @location.setter
    def location(self, value):
        self[PidLidLocation] = value

    @property
    def recurring(self):
        """Appointment is recurring."""
        return self._get_fast(self.store._pidlid_proptag(PidLidRecurring))

    @recurring.setter
    def recurring(self, value): # TODO update/invalidate cache
        # TODO cleanup on False?
        if value and not self.recurring:
            Recurrence._init(self)
        self[PidLidRecurring] = value

    @property
    def busystatus(self): # TODO deprecate and use busy_status instead?
        proptag = self.store._pidlid_proptag(PidLidBusyStatus)
        return FB_STATUS.get(self._get_fast(proptag))

    @busystatus.setter
    def busystatus(self, val):
        try:
            # props are identical
            self.create_prop('appointment:33285', STATUS_FB[val], PT_LONG)
        except KeyError:
            raise ArgumentError('invalid busy status: %s' % val)

    @property
    def recurrence(self):
        """Appointment :class:`recurrence <Recurrence>`."""
        try:
            return Recurrence(self)
        except NotFoundError:
            pass

    def occurrences(self, start=None, end=None, page_start=None,
                    page_limit=None, order=None):
        """Appointment :class:`occurrences <Occurrence>` (expanding
        recurring appointments).

        :param start: start from given date (optional)
        :param end: end at given date (optional)
        """

        pos = 0
        count = 0
        recurrence = self.recurrence
        if self.recurring and recurrence:
            for occurrence in self.recurrence.occurrences(start=start, end=end):
                if page_start is None or pos >= page_start:
                    yield occurrence
                    count += 1
                if page_limit is not None and count >= page_limit:
                    break
                pos += 1
        else:
            if (not start or self.end > start) and \
               (not end or self.start < end):
                start = max(self.start, start) if start else self.start
                end = min(self.end, end) if end else self.end
                yield Occurrence(self, start, end)

        if not recurrence and self.recurring:
            self.server.log.warn("Item '%s' has no recurrence ", self.entryid)
        if recurrence and not self.recurring:
            self.server.log.warn("Item '%s' has a recurrence but is not recurring", self.entryid)

    def occurrence(self, id_=None):
        if self.recurring:
            return self.recurrence.occurrence(id_)
        else:
            # TODO check if matches args
            return Occurrence(self)

    @property
    def reminder(self):
        """Is appointment reminder set."""
        return self.get(PidLidReminderSet, False)

    @reminder.setter
    def reminder(self, value): # TODO move to item, because common?
        self[PidLidReminderSet] = value

    @property
    def reminder_minutes(self):
        """Reminder minutes before appointment."""
        return self.get(PidLidReminderDelta)

    @reminder_minutes.setter
    def reminder_minutes(self, value):
        self[PidLidReminderDelta] = value

    @property
    def rrule(self): # TODO including timezone!
        if self.recurring: # TODO rrule for non-recurring makes sense?
            return self.recurrence.recurrences

    # TODO rrule setter!

    def attendees(self):
        """Appointment :class:`attendees <Attendee>`."""

        # Filter out organizer from all recipients
        restriction = Restriction(SOrRestriction([SNotRestriction(SExistRestriction(PR_RECIPIENT_FLAGS)),
                                                  SBitMaskRestriction(BMR_EQZ, PR_RECIPIENT_FLAGS, recipOrganizer)]))
        for row in self.table(PR_MESSAGE_RECIPIENTS, restriction=restriction):
            yield Attendee(self.server, row)

    def create_attendee(self, type_, address):
        """Create appointment :class:`attendee <Attendee>`.

        :param type_: attendee type (required, optional or resource)
        :param address: attendee address (str or :class:`address <Address>`)
        """
        # TODO move to Attendee class
        # TODO return Attendee instance

        reciptype = {
            'required': 1,
            'optional': 2,
            'resource': 3
        }[type_]

        names = []
        pr_addrtype, pr_dispname, pr_email, pr_entryid = \
            self._addr_props(address)
        names.append([
            SPropValue(PR_RECIPIENT_TYPE, reciptype),
            SPropValue(PR_DISPLAY_NAME_W, pr_dispname),
            SPropValue(PR_ADDRTYPE_W, str(pr_addrtype)),
            SPropValue(PR_EMAIL_ADDRESS_W, str(pr_email)),
            SPropValue(PR_ENTRYID, pr_entryid),
        ])
        self.mapiobj.ModifyRecipients(MODRECIP_ADD, names)

        self[PidLidAppointmentStateFlags] = ASF_MEETING
        self[PidLidResponseStatus] = respOrganized # TODO delegation?

        _utils._save(self.mapiobj)

    @property
    def response_requested(self):
        """Is appointment response requested."""
        return self.get(PR_RESPONSE_REQUESTED, False)

    @property
    def response_status(self):
        try:
            return RESPONSE_STATUS.get(self[PidLidResponseStatus])
        except NotFoundError:
            return 'None'

    @response_status.setter
    def response_status(self, val):
        try:
            # props are identical
            self.create_prop('appointment:33304', STATUS_RESPONSE[val], PT_LONG)
        except KeyError:
            raise ArgumentError('invalid response status: %s' % val)

    @property
    def icaluid(self):
        """Appointment iCal UID."""
        try:
            return _benc(self[PidLidGlobalObjectId])
        except NotFoundError:
            pass

    @property
    def eventid(self):
        # msgraph has both appointments and expanded appointments under
        # /events, so we need an identier which can be used for both.
        eid = _bdec(self.entryid)
        return _benc(b'\x00' + _utils.pack_short(len(eid)) + eid)

    @property
    def tzinfo(self):
        """Appointment timezone as datetime compatible tzinfo object."""
        tzdata = self._get_fast(self.store._pidlid_proptag(PidLidTimeZoneStruct))
        if tzdata:
            return _timezone.MAPITimezone(tzdata)

    @property
    def timezone(self):
        """Appointment timezone description."""
        return self.get(PidLidTimeZoneDescription)

    @timezone.setter
    def timezone(self, value):
        self[PidLidTimeZoneDescription] = str(value)
        self[PidLidTimeZoneStruct] = _timezone._timezone_struct(value)

    @property
    def replytime(self):
        return self.get(PidLidAppointmentReplyTime)

    @replytime.setter
    def replytime(self, value):
        self[PidLidAppointmentReplyTime] = value

    @property
    def replyname(self):
        return self.get(PidLidAppointmentReplyName)

    @replyname.setter
    def replyname(self, value):
        self[PidLidAppointmentReplyName] = value

    def accept(
        self,
        comment=None,
        tentative=False,
        respond=True,
        subject_prefix=None,
        proposed_new_time=None
    ):
        """Accept an appointment.

        Args:
            comment (str): accept with a comment. Default to None.
            tentative (bool): is it tentatively accepted or not. Default to False.
            respond (bool): respond needs to be sent to the organizer or not. Default to True.
            subject_prefix (str): respond subject prefix. Default to None.
            proposed_new_time (tuple): propose a new time. Default to None.
        """
        if tentative:
            self.busystatus = 'tentative'
            self.response_status = 'TentativelyAccepted'
        else:
            self.busystatus = 'busy'
            self.response_status = 'Accepted'

        self.replytime = datetime.now()
        self.replyname = self.store.user.fullname

        if respond:
            if tentative:
                if not subject_prefix:
                    subject_prefix = 'Tentatively acccepted'

                if proposed_new_time:
                    self.proposed_new_time_start = proposed_new_time[0]
                    self.proposed_new_time_end = proposed_new_time[1]

                message_class = 'IPM.Schedule.Meeting.Resp.Tent'
                self._respond(subject_prefix, message_class, comment, proposed_new_time)
            else:
                if not subject_prefix:
                    subject_prefix = 'Accepted'
                message_class = 'IPM.Schedule.Meeting.Resp.Pos'
                self._respond(subject_prefix, message_class, comment)

    def decline(self, comment=None, respond=True, subject_prefix='Declined'):
        self.response_status = 'Declined'
        self.replytime = datetime.now()
        self.replyname = self.store.user.fullname

        if respond:
            message_class = 'IPM.Schedule.Meeting.Resp.Neg'
            self._respond(subject_prefix, message_class, comment)

    # TODO merge with meetingrequest version
    def _respond(self, subject_prefix, message_class, comment=None, proposed_new_time=None):
        response = self.copy(self.store.outbox)
        response.message_class = message_class

        if proposed_new_time:
            response.proposed_new_time_start = proposed_new_time[0]
            response.proposed_new_time_end = proposed_new_time[1]

        response.subject = subject_prefix + ': ' + self.subject
        if comment:
            response.text = comment

        if not self.from_.email:
            self.server.log.debug("Item '%s' has no organiser cannot send meeting request response", self.entryid)
            return

        response.from_ = self.store.user # TODO slow?

        try:
            response.to = self.server.user(email=self.from_.email)
        except NotFoundError:
            response.to = self.from_.email

        response.send()

    @property
    def allow_new_time_proposals(self):
        """PidLidAppointmentCounterProposal getter.

        Returns:
            bool: allow new proposals to be suggested.
        """
        return self.get(PidLidAppointmentCounterProposal, False)

    @allow_new_time_proposals.setter
    def allow_new_time_proposals(self, value):
        """PidLidAppointmentCounterProposal setter.

        Args:
            value (bool): allow new proposals to be suggested.
        """
        self[PidLidAppointmentCounterProposal] = value

    def has_proposed_new_time(self):
        """Check the item has a proposed new time or not.

        Returns:
            bool: has proposed new time or not.
        """
        return self.proposed_new_time_start or self.proposed_new_time_end

    @property
    def proposed_new_time_start(self):
        """PidLidAppointmentStartWhole getter.

        Returns:
            datetime: proposed start datetime.
            None: if nothing set.
        """
        return self.get(PidLidAppointmentProposedStartWhole)

    @proposed_new_time_start.setter
    def proposed_new_time_start(self, new_datetime):
        """PidLidAppointmentStartWhole setter.

        Args:
            new_datetime (datetime): new datetime
        """
        self[PidLidAppointmentProposedStartWhole] = new_datetime

    @property
    def proposed_new_time_end(self):
        """PidLidAppointmentEndWhole getter.

        Returns:
            datetime: proposed end datetime.
            None: if nothing set.
        """
        return self.get(PidLidAppointmentProposedEndWhole)

    @proposed_new_time_end.setter
    def proposed_new_time_end(self, new_datetime):
        """PidLidAppointmentEndWhole setter.

        Args:
            new_datetime (datetime): new datetime
        """
        self[PidLidAppointmentProposedEndWhole] = new_datetime

    def cancel(self):
        self[PidLidAppointmentStateFlags] |= ASF_CANCELED

    @property
    def canceled(self):
        """Is appointment canceled."""
        return bool(self[PidLidAppointmentStateFlags] & ASF_CANCELED)

    @property
    def color(self): # property used by old clients
        """Appointment color (old clients)."""
        nr = self.get(PidLidAppointmentColor, 0)
        if nr != 0:
            return NR_COLOR[nr]

    @color.setter
    def color(self, value):
        try:
            self[PidLidAppointmentColor] = COLOR_NR[value]
        except KeyError:
            raise ArgumentError('invalid color: %r' % value)

    @property
    def onlinemeetingurl(self):
        try:
            return self.prop('public:OnlineMeetingExternalLink', proptype=PT_UNICODE).value
        except NotFoundError:
            return ''

    @onlinemeetingurl.setter
    def onlinemeetingurl(self, value):
        if not isinstance(value, str):
            raise ArgumentError('meeting url must be a string')
        self.create_prop('public:OnlineMeetingExternalLink', value, proptype=PT_UNICODE)



# SPDX-License-Identifier: AGPL-3.0-only
from MAPI import (TBL_BATCH, MAPI_UNICODE, MAPI_DISTLIST, FL_FULLSTRING,
                  FL_IGNORECASE, DT_REMOTE_MAILUSER, MAPI_MAILUSER, DT_GLOBAL,
                  RELOP_EQ, RELOP_NE, BOOKMARK_BEGINNING)
from MAPI.Defs import DEFINE_GUID
from MAPI.Tags import (PR_ENTRYID, PR_EMS_AB_CONTAINERID, PR_DISPLAY_TYPE,
                       PR_DISPLAY_NAME_W, PR_DISPLAY_NAME_A, PR_EMAIL_ADDRESS_A,
                       PR_EMAIL_ADDRESS_W, PR_OBJECT_TYPE)
from MAPI.Struct import (SPropertyRestriction, SOrRestriction,
                         SExistRestriction, SAndRestriction, SPropValue, SContentRestriction)

MUIDECSAB = DEFINE_GUID(0x50a921ac, 0xd340, 0x48ee, 0xb3, 0x19, 0xfb, 0xa7, 0x53, 0x30, 0x44, 0x25)

def GetGab(session):
    ab = session.OpenAddressBook(0, None, 0)
    root = ab.OpenEntry(None, None, 0)
    table = root.GetHierarchyTable(0)
    table.SetColumns([PR_ENTRYID], TBL_BATCH)
    restriction = SOrRestriction([SPropertyRestriction(RELOP_EQ, PR_DISPLAY_TYPE, SPropValue(PR_DISPLAY_TYPE, DT_GLOBAL)),
                                  SAndRestriction([SExistRestriction(PR_EMS_AB_CONTAINERID),
                                                   SPropertyRestriction(RELOP_EQ, PR_EMS_AB_CONTAINERID, SPropValue(PR_EMS_AB_CONTAINERID, 0))])])
    table.FindRow(restriction, BOOKMARK_BEGINNING, 0)
    eid = table.QueryRows(1, 0)[0][0].Value
    return ab.OpenEntry(eid, None, 0)

def GetCompanyList(session, flags = 0):
    if flags & MAPI_UNICODE:
        columns = [PR_DISPLAY_NAME_W]
    else:
        columns = [PR_DISPLAY_NAME_A]
    companies = []
    gab = GetGab(session)
    table = gab.GetHierarchyTable(0)
    table.SetColumns(columns, TBL_BATCH)
    while True:
        rows = table.QueryRows(50, 0)
        if len(rows) == 0:
            break
        companies.extend((row[0].Value for row in rows))
    return companies

def _GetAbObjectList(container, restriction, flags):
    if flags & MAPI_UNICODE:
        columns = [PR_EMAIL_ADDRESS_W]
    else:
        columns = [PR_EMAIL_ADDRESS_A]
    users = []
    table = container.GetContentsTable(0)
    table.SetColumns(columns, TBL_BATCH)
    if restriction:
        table.Restrict(restriction, TBL_BATCH)
    while True:
        rows = table.QueryRows(50, 0)
        if len(rows) == 0:
            break
        users.extend((row[0].Value for row in rows))
    return users

def GetAbObjectList(session, restriction = None, companyname = None, flags = 0):
    gab = GetGab(session)
    table = gab.GetHierarchyTable(0)
    if companyname is None or len(companyname) == 0:
        companyCount = table.GetRowCount(0)
        if companyCount == 0:
            users = _GetAbObjectList(gab, restriction, flags)
        else:
            table.SetColumns([PR_ENTRYID], TBL_BATCH)
            users = []
            for row in table.QueryRows(companyCount, 0):
                company = gab.OpenEntry(row[0].Value, None, 0)
                companyUsers = _GetAbObjectList(company, restriction, flags)
                users.extend(companyUsers)
    else:
        if isinstance(companyname, str):
            tag_dispname = PR_DISPLAY_NAME_W
        else:
            tag_dispname = PR_DISPLAY_NAME_A
        table.FindRow(SContentRestriction(FL_FULLSTRING|FL_IGNORECASE, tag_dispname, SPropValue(tag_dispname, companyname)), BOOKMARK_BEGINNING, 0)
        table.SetColumns([PR_ENTRYID], TBL_BATCH)
        row = table.QueryRows(1, 0)[0]
        users = _GetAbObjectList(gab.OpenEntry(row[0].Value, None, 0), restriction, flags)

    return users

def GetUserList(session, companyname = None, flags = 0):
    restriction = SAndRestriction([SPropertyRestriction(RELOP_EQ, PR_OBJECT_TYPE, SPropValue(PR_OBJECT_TYPE, MAPI_MAILUSER)),
                                   SPropertyRestriction(RELOP_NE, PR_DISPLAY_TYPE, SPropValue(PR_DISPLAY_TYPE, DT_REMOTE_MAILUSER))])
    return GetAbObjectList(session, restriction, companyname, flags)

def GetGroupList(session, companyname = None, flags = 0):
    restriction = SPropertyRestriction(RELOP_EQ, PR_OBJECT_TYPE, SPropValue(PR_OBJECT_TYPE, MAPI_DISTLIST))
    return GetAbObjectList(session, restriction, companyname, flags)



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import os

from .errors import ConfigError

from . import utils as _utils

KOPANO_CFG_DIR = os.getenv('KOPANO_CFG_DIR') or '/etc/kopano'

class ConfigOption:
    """Configuration Option class

    Abstraction for single configuration option.
    """
    def __init__(self, type_, **kwargs):
        self.type_ = type_
        self.kwargs = kwargs

    def parse(self, key, value):
        return getattr(self, 'parse_' + self.type_)(key, value)

    def parse_string(self, key, value):
        if self.kwargs.get('multiple'):
            values = value.split()
        else:
            values = [value]
        for value in values:
            # TODO moved to parse_path
            if self.kwargs.get('check_path') and not os.path.exists(value):
                raise ConfigError(
                    "%s: path '%s' does not exist" % (key, value))
            if (self.kwargs.get('options') is not None and \
                value not in self.kwargs.get('options')):
                raise ConfigError(
                    "%s: '%s' is not a legal value" % (key, value))
        if self.kwargs.get('multiple'):
            return values
        else:
            return values[0]

    def parse_path(self, key, value):
        if self.kwargs.get('check', True) and not os.path.exists(value):
            raise ConfigError("%s: path '%s' does not exist" % (key, value))
        return value

    def parse_integer(self, key, value):
        if (self.kwargs.get('options') is not None and \
            int(value) not in self.kwargs.get('options')):
            raise ConfigError("%s: '%s' is not a legal value" % (key, value))
        if self.kwargs.get('multiple'):
            base = self.kwargs.get('base', 10)
            return [int(x, base=base) for x in value.split()]
        try:
            return int(value, base=self.kwargs.get('base', 10))
        except ValueError:
            raise ConfigError("%s: '%s' is not a legal value" % (key, value))

    def parse_boolean(self, key, value):
        return {
            'no': False,
            'yes': True,
            '0': False,
            '1': True,
            'false': False,
            'true': True
        }[value]

    def parse_size(self, key, value):
        return _utils.human_to_bytes(value)

class Config:
    """Configuration class

    Abstraction for kopano-style configation files.
    """
    def __init__(self, config=None, service=None, options=None, filename=None,
            log=None):
        self.config = config
        self.service = service
        self.warnings = []
        self.errors = []
        self.info = []
        if filename:
            pass
        elif options and getattr(options, 'config_file', None):
            filename = options.config_file
        elif service:
            filename = '%s/%s.cfg' % (KOPANO_CFG_DIR, service)
        self.data = {}
        if self.config is not None:
            for key, val in self.config.items():
                if 'default' in val.kwargs:
                    self.data[key] = val.kwargs.get('default')
        try:
            fh = open(filename, "r")
        except:
            msg = "could not open config file %s, using defaults" % filename
            self.info.append(msg)
        else:
            self._parse_config(fh)

        if self.config is not None:
            for key, val in self.config.items():
                if key not in self.data and val.type_ != 'ignore':
                    msg = "%s: missing in config file" % key
                    if service: # TODO merge
                        self.errors.append(msg)
                    else:
                        raise ConfigError(msg)

    def _parse_config(self, fh):
        for line in fh:
            line = line.strip()
            if line.startswith('#'):
                continue
            pos = line.find('=')
            if pos == -1:
                continue
            key = line[:pos].strip()
            value = line[pos + 1:].strip()
            if self.config is None:
                self.data[key] = value
            elif key in self.config:
                if self.config[key].type_ == 'ignore':
                    self.data[key] = None
                    self.warnings.append('%s: config option ignored' % key)
                else:
                    try:
                        self.data[key] = self.config[key].parse(key, value)
                    except ConfigError as e:
                        if self.service:
                            self.errors.append(str(e))
                        else:
                            raise
            else:
                msg = "%s: unknown config option" % key
                if self.service:
                    self.warnings.append(msg)
                else:
                    raise ConfigError(msg)

    @staticmethod
    def string(**kwargs):
        return ConfigOption(type_='string', **kwargs)

    @staticmethod
    def path(**kwargs):
        return ConfigOption(type_='path', **kwargs)

    @staticmethod
    def boolean(**kwargs):
        return ConfigOption(type_='boolean', **kwargs)

    @staticmethod
    def integer(**kwargs):
        return ConfigOption(type_='integer', **kwargs)

    @staticmethod
    def size(**kwargs):
        return ConfigOption(type_='size', **kwargs)

    @staticmethod
    def ignore(**kwargs):
        return ConfigOption(type_='ignore', **kwargs)

    def get(self, x, default=None):
        return self.data.get(x, default)

    def __getitem__(self, x):
        return self.data[x]

# TODO to log.py
LOG_LEVELS = [str(i) for i in range(7)] + \
    ['info', 'debug', 'warning', 'error', 'critical']

CONFIG = {
    'log_method': Config.string(options=['file', 'syslog'], default='file'),
    'log_level': Config.string(options=LOG_LEVELS, default='warning'),
    'log_file': Config.string(default="-"),
    'log_timestamp': Config.boolean(default="true"),
    'pid_file': Config.string(default=None),
    'run_as_user': Config.string(default=None),
    'run_as_group': Config.string(default=None),
    'server_socket': Config.string(default=None),
    'sslkey_file': Config.path(default=None),
    'sslkey_pass': Config.string(default=None),
    'worker_processes': Config.integer(default=1),
}



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import codecs
import collections
import mailbox
import sys
import time
from xml.etree import ElementTree

import icalmapi

from MAPI import (
    MAPI_MODIFY, MAPI_ASSOCIATED, ROW_ADD,
    RELOP_GT, RELOP_LT, RELOP_EQ, MAPI_CREATE,
    DEL_ASSOCIATED, DEL_FOLDERS, DEL_MESSAGES, COPY_SUBFOLDERS,
    BOOKMARK_BEGINNING, ROW_REMOVE, MESSAGE_MOVE, FOLDER_MOVE,
    FOLDER_GENERIC, MAPI_UNICODE, FL_SUBSTRING, FL_IGNORECASE,
    SEARCH_RECURSIVE, SEARCH_REBUILD, PT_MV_BINARY, PT_BINARY,
    MAPI_DEFERRED_ERRORS, FL_PREFIX
)
from MAPI.Tags import (
    PR_ENTRYID, IID_IMAPIFolder, SHOW_SOFT_DELETES, PR_SOURCE_KEY,
    PR_PARENT_ENTRYID, PR_EC_HIERARCHYID, PR_FOLDER_CHILD_COUNT,
    PR_DISPLAY_NAME_W, PR_CONTAINER_CLASS_W, PR_CONTENT_UNREAD,
    PR_MESSAGE_DELIVERY_TIME, MNID_ID, PT_SYSTIME, PT_BOOLEAN,
    DELETE_HARD_DELETE, PR_MESSAGE_SIZE, PR_ACL_TABLE, PR_MESSAGE_CLASS_W,
    PR_MEMBER_ID, PR_RULES_TABLE, IID_IExchangeModifyTable,
    IID_IMAPITable, PR_CONTAINER_CONTENTS, PR_RULES_DATA, PR_STORE_ENTRYID,
    PR_FOLDER_ASSOCIATED_CONTENTS, PR_CONTAINER_HIERARCHY,
    PR_SUBJECT_W, PR_BODY_W, PR_DISPLAY_TO_W, PR_CREATION_TIME,
    CONVENIENT_DEPTH, PR_CONTENT_COUNT, PR_ASSOC_CONTENT_COUNT,
    PR_DELETED_MSG_COUNT, PR_LAST_MODIFICATION_TIME, PR_MESSAGE_ATTACHMENTS,
    PR_EC_PUBLIC_IPM_SUBTREE_ENTRYID, PR_CHANGE_KEY, PR_EXCEPTION_STARTTIME,
    PR_EXCEPTION_ENDTIME, PR_RULE_ID, PR_RULE_STATE, ST_ENABLED,
    PR_RULE_PROVIDER_DATA, PR_RULE_SEQUENCE, PR_RULE_NAME_W,
    PR_RULE_CONDITION, PT_LONG, PR_ATTR_HIDDEN,
)
from MAPI.Defs import (
    HrGetOneProp, CHANGE_PROP_TYPE
)
from MAPI.Struct import (
    MAPIErrorNoAccess, MAPIErrorNotFound, MAPIErrorNoSupport,
    MAPIErrorInvalidEntryid, MAPIErrorCollision, SPropValue,
    MAPINAMEID, SOrRestriction, SAndRestriction, SPropertyRestriction,
    SSubRestriction, SContentRestriction, ROWENTRY
)
from MAPI.Time import unixtime

from .properties import Properties
from .permission import Permission, _permissions_dumps, _permissions_loads
from .restriction import Restriction
from .recurrence import Occurrence
from .rule import Rule
from .table import Table
from .property_ import Property
from .defs import (
    PSETID_Appointment, UNESCAPED_SLASH_RE,
    ENGLISH_FOLDER_MAP, NAME_RIGHT, NAMED_PROPS_ARCHIVER
)
from .errors import NotFoundError, ArgumentError, DuplicateError
from .query import _query_to_restriction

from .compat import (
    bdec as _bdec, benc as _benc
)

try:
    from . import log as _log
except ImportError: # pragma: no cover
    _log = sys.modules[__package__ + '.log']
try:
    from . import user as _user
except ImportError: # pragma: no cover
    _user = sys.modules[__package__ + '.user']
try:
    from . import store as _store
except ImportError: # pragma: no cover
    _store = sys.modules[__package__ + '.store']
try:
    from . import item as _item
except ImportError: # pragma: no cover
    _item = sys.modules[__package__ + '.item']
try:
    from . import utils as _utils
except ImportError: # pragma: no cover
    _utils = sys.modules[__package__ + '.utils']
try:
    from . import ics as _ics
except ImportError: # pragma: no cover
    _ics = sys.modules[__package__ + '.ics']
try:
    from . import notification as _notification
except ImportError: # pragma: no cover
    _notification = sys.modules[__package__ + '.notification']

def _unbase64(s):
    return codecs.decode(codecs.encode(s, 'ascii'), 'base64')

def _base64(s):
    return codecs.decode(codecs.encode(s, 'base64').strip(), 'ascii')

# TODO generalize, autogenerate basic item getters/setters?
PROPMAP = {
    'subject': PR_SUBJECT_W,
    'received': PR_MESSAGE_DELIVERY_TIME,
    'created': PR_CREATION_TIME,
}

class Folder(Properties):
    """Folder class

    Abstraction for collections of :class:`items <Item>`. For example,
    folders, calendars and contact folders.

    Includes all functionality from :class:`Properties`.
    """

    def __init__(self, store=None, entryid=None, associated=False,
                 deleted=False, mapiobj=None, _check_mapiobj=True, cache={}):
        if store:
            self.store = store
            self.server = store.server
        # TODO else, determine store dynamically?
        if mapiobj:
            self._mapiobj = mapiobj
            self._entryid = HrGetOneProp(self.mapiobj, PR_ENTRYID).Value
        elif entryid:
            try:
                self._entryid = _bdec(entryid)
            except:
                raise ArgumentError('invalid entryid: %r' % entryid)

        self._content_flag = MAPI_ASSOCIATED if associated else \
            (SHOW_SOFT_DELETES if deleted else 0)
        self._sourcekey = None
        self._mapiobj = None

        if _check_mapiobj: # raise error for specific key
            self.mapiobj

        self._cache = cache
        self._iter = None

    @property
    def mapiobj(self):
        """Underlying MAPI object."""
        if self._mapiobj:
            return self._mapiobj

        try:
            self._mapiobj = self.store.mapiobj.OpenEntry(self._entryid,
                IID_IMAPIFolder, MAPI_MODIFY)
        except (MAPIErrorNotFound, MAPIErrorInvalidEntryid):
            try:
                self._mapiobj = self.store.mapiobj.OpenEntry(self._entryid,
                    IID_IMAPIFolder, MAPI_MODIFY | SHOW_SOFT_DELETES)
            # TODO check too late?
            except (MAPIErrorNotFound, MAPIErrorInvalidEntryid):
                raise NotFoundError(
                    "no folder with entryid '%s'" % _benc(self._entryid))
        except MAPIErrorNoAccess: # TODO
            self._mapiobj = self.store.mapiobj.OpenEntry(self._entryid,
                IID_IMAPIFolder, 0)

        return self._mapiobj

    @mapiobj.setter
    def mapiobj(self, mapiobj):
        self._mapiobj = mapiobj

    @property
    def entryid(self):
        """Folder entryid."""
        return _benc(self._entryid)

    @property
    def guid(self):
        return self.entryid[56:88] # TODO assumes hex

    @property
    def sourcekey(self):
        """Folder sourcekey."""
        if not self._sourcekey:
            self._sourcekey = \
                _benc(HrGetOneProp(self.mapiobj, PR_SOURCE_KEY).Value)
        return self._sourcekey

    @property
    def parent(self):
        """:class:`Parent <Folder>` folder"""
        parent_eid = self._get_fast(PR_PARENT_ENTRYID)
        if parent_eid:
            parent_eid = _benc(parent_eid)
            if parent_eid != self.entryid: # root parent is itself
                return Folder(self.store, parent_eid, _check_mapiobj=False)

    @property
    def hierarchyid(self):
        """Folder hierarchy (SQL) id."""

        return self.prop(PR_EC_HIERARCHYID).value

    @property
    def subfolder_count(self):
        """Direct subfolder count."""
        return self.prop(PR_FOLDER_CHILD_COUNT).value

    @property
    def subfolder_count_recursive(self):
        """Subfolder count (recursive)."""
        # TODO faster?
        flags = MAPI_UNICODE | self._content_flag | CONVENIENT_DEPTH | \
            MAPI_DEFERRED_ERRORS
        mapitable = self.mapiobj.GetHierarchyTable(flags)

        table = Table(
            self.server,
            self.mapiobj,
            mapitable,
            PR_CONTAINER_HIERARCHY,
        )

        return table.count

    @property
    def name(self):
        """Folder name."""
        name = self._get_fast(PR_DISPLAY_NAME_W)

        if name is not None:
            return name.replace('/', '\\/')
        else:
            # root folder doesn't have PR_DISPLAY_NAME_W
            if self.entryid == self.store.root.entryid:
                return 'ROOT'
            else:
                return ''

    @property
    def path(self):
        """Folder path."""
        names = []
        parent = self
        subtree_entryid = self.store.subtree.entryid
        root_entryid = self.store.root.entryid
        # parent entryids coming from a table are not converted like normal
        # parent entryids for certains public store folders (via getprops,
        # or store PR_IPM_SUBTREE_ENTRYID), so we match with this property,
        # which is also not converted
        pub_entryid = \
            _benc(self.store.get(PR_EC_PUBLIC_IPM_SUBTREE_ENTRYID, b''))

        while parent:
            parent_eid = parent.entryid
            if parent_eid in (subtree_entryid, pub_entryid):
                return '/'.join(reversed(names)) if names else None
            elif parent_eid == root_entryid:
                return
            else:
                names.append(parent.name)
                parent = parent.parent

    @name.setter
    def name(self, name):
        self.mapiobj.SetProps([SPropValue(PR_DISPLAY_NAME_W, str(name))])
        _utils._save(self.mapiobj)

    @property
    def container_class(self): # TODO return '' by default?
        """
        Describes the type of items a folder holds. Possible values:

        * IPF.Appointment
        * IPF.Contact
        * IPF.Journal
        * IPF.Note
        * IPF.StickyNote
        * IPF.Task

        https://msdn.microsoft.com/en-us/library/aa125193(v=exchg.65).aspx
        """
        try:
            return self.prop(PR_CONTAINER_CLASS_W).value
        except NotFoundError:
            pass

    @container_class.setter
    def container_class(self, value):
        if value is None:
            return
        prop = SPropValue(PR_CONTAINER_CLASS_W, str(value))
        self.mapiobj.SetProps([prop])
        _utils._save(self.mapiobj)

    @property
    def hidden(self):
        try:
            return self.prop(PR_ATTR_HIDDEN).value
        except NotFoundError:
            return False

    @hidden.setter
    def hidden(self, value):
        prop = SPropValue(PR_ATTR_HIDDEN, value)
        self.mapiobj.SetProps([prop])
        _utils._save(self.mapiobj)

    @property
    def unread(self):
        """Unread item count."""
        return self._get_fast(PR_CONTENT_UNREAD)

    def item(self, entryid=None, sourcekey=None):
        """ Return :class:`Item` with given entryid or sourcekey

        :param entryid: item entryid (optional)
        :param sourcekey: item sourcekey (optional)
        """
        if entryid is not None:
            eid = _utils._bdec_eid(entryid)

        elif sourcekey is not None: # TODO 1 SQL per row with empty cache!
            try:
                restriction = SPropertyRestriction(RELOP_EQ, PR_SOURCE_KEY,
                    SPropValue(PR_SOURCE_KEY, _bdec(sourcekey)))
            except:
                raise ArgumentError("invalid sourcekey: %r" % sourcekey)
            table = self.mapiobj.GetContentsTable(MAPI_DEFERRED_ERRORS)
            table.SetColumns([PR_ENTRYID, PR_SOURCE_KEY], 0)
            table.Restrict(restriction, 0)
            rows = list(table.QueryRows(2147483647, 0))
            if not rows:
                raise NotFoundError("no item with sourcekey '%s'" % sourcekey)
            eid = rows[0][0].Value

        else:
            raise ArgumentError("no entryid or sourcekey specified")

        # open message with entryid
        try:
            mapiobj = _utils.openentry_raw(self.store.mapiobj, eid,
                self._content_flag)
        except MAPIErrorNotFound:
            if sourcekey is not None:
                raise NotFoundError("no item with sourcekey '%s'" % sourcekey) from None
            else:
                raise NotFoundError("no item with entryid '%s'" % entryid) from None
        except MAPIErrorInvalidEntryid:
            raise ArgumentError("invalid entryid: %r" % entryid) from None

        item = _item.Item(self, mapiobj=mapiobj)
        return item

    def items(self, restriction=None, page_start=None, page_limit=None,
              order=None, query=None):
        """Return all :class:`items <Item>` in folder, reverse sorted on
        received date.

        :param restriction: apply :class:`restriction <Restriction>`
        :param order: order by (limited set of) attributes, e.g. 'subject',
            '-subject' (reverse order), or ('subject', '-received').
        :param page_start: skip this many items from the start
        :param page_limit: return up to this many items
        :param query: use search query
        """
        # TODO determine preload columns dynamically, based on usage pattern
        columns = [
            PR_ENTRYID,
            PR_MESSAGE_DELIVERY_TIME,
            PR_SUBJECT_W, # watch out: table unicode data is max 255 chars
            PR_LAST_MODIFICATION_TIME,
            PR_CHANGE_KEY,
            PR_MESSAGE_CLASS_W,
            PR_BODY_W, # body preview, not entire body!
            PR_SOURCE_KEY,
        ]

        if query is not None:
            if self.container_class == 'IPF.Contact':
                type_ = 'contact'
            else:
                type_ = 'message'
            restriction = _query_to_restriction(query, type_, self.store)

        try:
            table = Table(
                self.server,
                self.mapiobj,
                self.mapiobj.GetContentsTable(self._content_flag | \
                    MAPI_DEFERRED_ERRORS),
                PR_CONTAINER_CONTENTS,
                columns=columns,
                restriction=restriction,
            )
        except MAPIErrorNoSupport:
            return

        # TODO MAPI has more than just ascend/descend
        if order is None:
            order = '-received'
        if not isinstance(order, tuple):
            order = (order,)
        sorttags = []
        for term in order:
            if term.startswith('-'):
                sorttags.append(-PROPMAP[term[1:]])
            else:
                sorttags.append(PROPMAP[term])
        table.sort(tuple(sorttags))

        for row in table.rows(page_start=page_start, page_limit=page_limit):
            item = _item.Item(
                self,
                entryid=row[0].value,
                content_flag=self._content_flag,
                cache=dict(zip(columns, row))
            )
            yield item

    def occurrences(self, start=None, end=None, page_start=None,
                    page_limit=None, order=None):
        """For applicable folder types (e.g., calendars), return
        all :class:`occurrences <Occurrence>`.

        :param start: start time (optional)
        :param end: end time (optional)
        """
        count = 0
        pos = 0
        if start and end:
            startstamp = time.mktime(start.timetuple())
            endstamp = time.mktime(end.timetuple())

            # TODO use shortcuts and default type (database) to avoid
            # MAPI snake wrestling
            NAMED_PROPS = [MAPINAMEID(PSETID_Appointment, MNID_ID, x)
                           for x in (33285, 33293, 33294, 33315, 33301, 33333, 33334, 33331, 33302)]
            ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS, MAPI_CREATE)
            busystatus = ids[0] | PT_LONG
            startdate = ids[1] | PT_SYSTIME
            enddate = ids[2] | PT_SYSTIME
            recurring = ids[3] | PT_BOOLEAN
            all_day = ids[4] | PT_BOOLEAN
            clip_start = ids[5] | PT_SYSTIME
            clip_end = ids[6] | PT_SYSTIME
            tzinfo = ids[7] | PT_BINARY
            blob = ids[8] | PT_BINARY

            restriction = SOrRestriction([
                # non-recurring: normal start/end
                SAndRestriction([
                    SPropertyRestriction(RELOP_GT, enddate, SPropValue(enddate, unixtime(startstamp))),
                    SPropertyRestriction(RELOP_LT, startdate, SPropValue(startdate, unixtime(endstamp))),
                ]),
                # recurring: range start/end
                SAndRestriction([
                    SPropertyRestriction(RELOP_GT, clip_end, SPropValue(clip_end, unixtime(startstamp))),
                    SPropertyRestriction(RELOP_LT, clip_start, SPropValue(clip_start, unixtime(endstamp))),
                ]),
                # exceptions: exception start/end in attachment
                SAndRestriction([
                    SPropertyRestriction(RELOP_EQ, recurring, SPropValue(recurring, True)),
                    SSubRestriction(
                        PR_MESSAGE_ATTACHMENTS,
                        SAndRestriction([
                            SPropertyRestriction(RELOP_LT,
                                PR_EXCEPTION_STARTTIME,
                                SPropValue(PR_EXCEPTION_STARTTIME, unixtime(endstamp))),
                            SPropertyRestriction(RELOP_GT, PR_EXCEPTION_ENDTIME, SPropValue(PR_EXCEPTION_ENDTIME, unixtime(startstamp))),
                        ])
                    )
                ])
            ])

            # Restrict occurences to real calendar items.
            restriction = SAndRestriction([
                SContentRestriction(FL_PREFIX, PR_MESSAGE_CLASS_W, SPropValue(PR_MESSAGE_CLASS_W, 'IPM.Appointment')),
                restriction,
            ])

            columns = [
                PR_ENTRYID,
                PR_SUBJECT_W,  # watch out: table unicode data is max 255 chars
                PR_LAST_MODIFICATION_TIME,
                PR_CHANGE_KEY,
                startdate,
                enddate,
                recurring,
                all_day,
                busystatus,
                tzinfo,
                blob,  # watch out: can be larger than 255 chars.
            ]

            table = Table(
                self.server,
                self.mapiobj,
                self.mapiobj.GetContentsTable(MAPI_DEFERRED_ERRORS),
                PR_CONTAINER_CONTENTS,
                columns=columns,
            )
            table.mapitable.Restrict(restriction, 0)
            for row in table.rows():
                item = _item.Item(
                    self,
                    entryid=row[0].value,
                    content_flag=self._content_flag,
                    cache=dict(zip(columns, row))
                )
                for occurrence in item.occurrences(start, end):
                    if page_start is None or pos >= page_start:
                        yield occurrence
                        count += 1
                    if page_limit is not None and count >= page_limit:
                        break
                    pos += 1
                if page_limit is not None and count >= page_limit:
                    break

        else:
            for item in self:
                for occurrence in item.occurrences(start, end):
                    if page_start is None or pos >= page_start:
                        yield occurrence
                        count += 1
                    if page_limit is not None and count >= page_limit:
                        break
                    pos += 1
                if page_limit is not None and count >= page_limit:
                    break

    def create_item(self, eml=None, ics=None, vcf=None, load=None, loads=None,
            attachments=True, save=True, **kwargs): # TODO associated
        """Create :class:`item <Item>`.

        :param eml: pass eml data/file (optional)
        :param ics: pass iCal data/file (optional)
        :param vcf: pass vcf data/file (optional)
        :param load: pass serialized item file (optional)
        :param loads: pass serialized item data (optional)
        """
        read = kwargs.get('read') # needs to be set before first save
        item = _item.Item(self, eml=eml, ics=ics, vcf=vcf, load=load,
            loads=loads, attachments=attachments, create=True, save=save,
            read=read)
        for key, val in kwargs.items():
            if key != 'read':
                setattr(item, key, val)
        return item

    # TODO: always hard delete or but we should also provide 'softdelete' which
    # moves the item to the wastebasket
    def empty(self, recurse=True, associated=False):
        """Delete folder contents (items and subfolders)

        :param recurse: delete subfolders (default True)
        :param associated: delete associated contents (default False)
        """
        if recurse:
            flags = DELETE_HARD_DELETE
            if associated:
                flags |= DEL_ASSOCIATED
            self.mapiobj.EmptyFolder(0, None, flags)
        else:
            self.delete(self.items()) # TODO look at associated flag!

    @property
    def size(self): # TODO bit slow perhaps? :P
        """Folder storage size."""
        try:
            table = Table(
                self.server,
                self.mapiobj,
                self.mapiobj.GetContentsTable(self._content_flag | \
                    MAPI_DEFERRED_ERRORS),
                PR_CONTAINER_CONTENTS,
                columns=[PR_MESSAGE_SIZE],
            )
        except MAPIErrorNoSupport:
            return 0

        table.mapitable.SeekRow(BOOKMARK_BEGINNING, 0)
        size = 0
        for row in table.rows():
            size += row[0].value
        return size

    @property
    def count(self):
        """Folder item count."""
        # reopen to get up-to-date counters..
        # TODO use flags when opening folder?
        folder = self.store.folder(entryid=self.entryid)

        if self._content_flag == 0:
            proptag = PR_CONTENT_COUNT
        elif self._content_flag == MAPI_ASSOCIATED:
            proptag = PR_ASSOC_CONTENT_COUNT
        elif self._content_flag == SHOW_SOFT_DELETES:
            proptag = PR_DELETED_MSG_COUNT
        return folder.get(proptag, 0)

    def recount(self):
        self.server.sa.ResetFolderCount(_bdec(self.entryid))

    def _get_entryids(self, items):
        item_entryids = [_bdec(item.entryid)
            for item in items if isinstance(item, _item.Item)]
        folder_entryids = [_bdec(item.entryid)
            for item in items if isinstance(item, Folder)]
        perms = [item for item in items if isinstance(item, Permission)]
        props = [item for item in items if isinstance(item, Property)]
        occs = [item for item in items if isinstance(item, Occurrence)]
        rules = [item for item in items if isinstance(item, Rule)]
        return item_entryids, folder_entryids, perms, props, occs, rules

    def delete(self, objects, soft=False): # TODO associated
        """Delete items, subfolders, properties or permissions from folder.

        :param objects: The object(s) to delete
        :param soft: soft-delete items/folders (default False)
        """
        objects = _utils.arg_objects(
            objects,
            (_item.Item, Folder, Permission, Property, Occurrence, Rule),
            'Folder.delete'
        )
        item_entryids, folder_entryids, perms, props, occs, rules = \
            self._get_entryids(objects)
        if item_entryids:
            if soft:
                self.mapiobj.DeleteMessages(item_entryids, 0, None, 0)
            else:
                self.mapiobj.DeleteMessages(item_entryids, 0, None,
                    DELETE_HARD_DELETE)
        for entryid in folder_entryids:
            if soft:
                self.mapiobj.DeleteFolder(entryid, 0, None,
                    DEL_FOLDERS | DEL_MESSAGES)
            else:
                self.mapiobj.DeleteFolder(entryid, 0, None,
                    DEL_FOLDERS | DEL_MESSAGES | DELETE_HARD_DELETE)
        for perm in perms:
            acl_table = self.mapiobj.OpenProperty(PR_ACL_TABLE,
                IID_IExchangeModifyTable, 0, 0)
            acl_table.ModifyTable(0, [ROWENTRY(ROW_REMOVE,
                [SPropValue(PR_MEMBER_ID, perm.mapirow[PR_MEMBER_ID])])])
        for rule in rules:
            rule_table = self.mapiobj.OpenProperty(PR_RULES_TABLE,
                IID_IExchangeModifyTable, MAPI_UNICODE, 0)
            rule_table.ModifyTable(0, [ROWENTRY(ROW_REMOVE,
                [SPropValue(PR_RULE_ID, rule.mapirow[PR_RULE_ID])])])

        if props:
            self.mapiobj.DeleteProps([prop.proptag for prop in props])
        for occ in occs:
            if occ.item.recurring:
                occ.item.delete(occ)
            else:
                self.delete(occ.item)

    def copy(self, objects, folder, _delete=False):
        """Copy items or subfolders to folder.

        :param objects: The items or subfolders to copy
        :param folder: The target folder
        """
        objects = _utils.arg_objects(
            objects, (_item.Item, Folder), 'Folder.copy')
        item_entryids, folder_entryids, _, _, _, _ = \
            self._get_entryids(objects)
        # TODO copy/move perms?? TODO error for perms/props
        if item_entryids:
            self.mapiobj.CopyMessages(item_entryids, IID_IMAPIFolder,
                folder.mapiobj, 0, None, (MESSAGE_MOVE if _delete else 0))
        for entryid in folder_entryids:
            self.mapiobj.CopyFolder(entryid, IID_IMAPIFolder, folder.mapiobj,
                None, 0, None,
                COPY_SUBFOLDERS | (FOLDER_MOVE if _delete else 0))

    def move(self, objects, folder):
        """Move items or subfolders to folder

        :param objects: The item(s) or subfolder(s) to move
        :param folder: The target folder
        """
        self.copy(objects, folder, _delete=True)

    def folder(self, path=None, entryid=None, recurse=False, create=False):
        """Return :class:`Folder` with given path or entryid

        :param path: Folder path (optional)
        :param entryid: Folder entryid (optional)
        :param create: Create folder if it doesn't exist (default False)
        """
        if entryid is not None:
            try:
                return Folder(self.store, entryid)
            except (MAPIErrorInvalidEntryid, MAPIErrorNotFound, TypeError):
                raise NotFoundError('no folder with entryid "%s"' % entryid)
        elif path is None:
            raise ArgumentError('missing argument to identify folder')

        if path is None:
            raise ArgumentError('no path or entryid specified')
        # TODO MAPI folders may contain '/' (and '\') in their names..
        if '/' in path.replace('\\/', ''):
            subfolder = self
            for name in UNESCAPED_SLASH_RE.split(path):
                subfolder = subfolder.folder(name, create=create,
                    recurse=False)
            return subfolder
        # TODO depth==0?
        if self == self.store.subtree and path in ENGLISH_FOLDER_MAP:
            f = getattr(self.store, ENGLISH_FOLDER_MAP[path], None)
            if f:
                path = f.name

        if create:
            name = path.replace('\\/', '/')
            try:
                mapifolder = self.mapiobj.CreateFolder(FOLDER_GENERIC,
                    str(name), '', None, MAPI_UNICODE)
                return Folder(self.store, _benc(HrGetOneProp(mapifolder,
                    PR_ENTRYID).Value))
            except MAPIErrorCollision:
                pass

        name = path.replace('\\/', '/')
        restriction = Restriction(SPropertyRestriction(RELOP_EQ,
            PR_DISPLAY_NAME_W, SPropValue(PR_DISPLAY_NAME_W, str(name))))

        folders = list(self.folders(recurse=recurse, restriction=restriction))
        if not folders:
            raise NotFoundError("no such folder: '%s'" % path)

        return folders[0]

    def get_folder(self, path=None, entryid=None):
        """Return :class:`folder <Folder>` with given name/entryid or
        *None* if not found

        :param path: Folder path (optional)
        :param entryid: Folder entryid (optional)
        """
        try:
            return self.folder(path, entryid=entryid)
        except NotFoundError:
            pass

    def folders(self, recurse=True, restriction=None, page_start=None,
                page_limit=None, order=None):
        """ Return all :class:`sub-folders <Folder>` in folder

        :param recurse: include all sub-folders
        :param restriction: apply :class:`restriction <Restriction>`
        :param page_start: skip this many items from the start
        :param page_limit: return up to this many items
        """
        # TODO implement order arg

        columns = [
            PR_ENTRYID,
            PR_PARENT_ENTRYID,
            PR_DISPLAY_NAME_W,
            PR_LAST_MODIFICATION_TIME,
            PR_CONTENT_UNREAD,
        ]

        flags = MAPI_UNICODE | self._content_flag
        if recurse:
            flags |= CONVENIENT_DEPTH | MAPI_DEFERRED_ERRORS

        try:
            mapitable = self.mapiobj.GetHierarchyTable(flags)

            table = Table(
                self.server,
                self.mapiobj,
                mapitable,
                PR_CONTAINER_HIERARCHY,
                columns=columns,
                restriction=restriction,
            )
        except MAPIErrorNoSupport: # TODO webapp search folder?
            return

        # determine all folders
        folders = {}
        names = {}
        children = collections.defaultdict(list)

        for row in table.rows(page_start=page_start, page_limit=page_limit):
            folder = Folder(
                self.store,
                _benc(row[0].value),
                _check_mapiobj=False,
                cache=dict(zip(columns, row))
            )

            folders[_benc(row[0].value)] = folder, _benc(row[1].value)
            names[_benc(row[0].value)] = row[2].value
            children[_benc(row[1].value)].append((_benc(row[0].value), folder))

        # yield depth-first TODO improve server?
        def folders_recursive(fs, depth=0):
            for feid, f in sorted(fs, key=lambda data: names[data[0]]):
                f.depth = depth
                yield f
                for f in folders_recursive(children[feid], depth + 1):
                    yield f

        rootfolders = []
        for eid, (folder, parenteid) in folders.items():
            if parenteid not in folders:
                rootfolders.append((eid, folder))
        for f in folders_recursive(rootfolders):
            yield f

            # SHOW_SOFT_DELETES filters out subfolders of soft-deleted folders
            # TODO slow
            if self._content_flag == SHOW_SOFT_DELETES:
                for g in f.folders():
                    g.depth += f.depth + 1
                    yield g

    def create_folder(self, path=None, **kwargs):
        """Create folder with given (relative) path

        :param path: Folder path
        """
        if path is None:
            path = kwargs.get('path', kwargs.get('name'))
        if self.get_folder(path):
            raise DuplicateError("folder '%s' already exists" % path)
        folder = self.folder(path, create=True)
        for key, val in kwargs.items():
            setattr(folder, key, val)
        return folder

    def rules(self):
        """Return all folder :class:`rules <Rule>`."""
        rule_table = self.mapiobj.OpenProperty(PR_RULES_TABLE,
            IID_IExchangeModifyTable, MAPI_UNICODE, 0)
        table = Table(
            self.server,
            self.mapiobj,
            rule_table.GetTable(0),
            PR_RULES_TABLE,
        )
        for row in table.dict_rows():
            yield Rule(row, rule_table)

    def create_rule(self, name=None, restriction=None):
        """Create rule

        :param name: Rule name (optional)
        :param restriction: Rule :class:`restriction <Restriction>` (optional)
        """
        if name:
            name = str(name)

        propid = len(list(self.rules()))+1
        row = [
            SPropValue(PR_RULE_STATE, ST_ENABLED), # TODO configurable
            SPropValue(PR_RULE_PROVIDER_DATA,
                codecs.decode('010000000000000074da402772c2e440', 'hex')),
            SPropValue(PR_RULE_SEQUENCE, propid), # TODO max+1?
            SPropValue(0x6681001f, 'RuleOrganizer'), # TODO name
            SPropValue(PR_RULE_ID, propid), # TODO max+1
        ]
        if name:
            row.append(SPropValue(PR_RULE_NAME_W, name))
        if restriction:
            row.append(SPropValue(PR_RULE_CONDITION, restriction.mapiobj))

        table = self.mapiobj.OpenProperty(PR_RULES_TABLE,
            IID_IExchangeModifyTable, MAPI_UNICODE, 0)
        table.ModifyTable(0, [ROWENTRY(ROW_ADD, row)])

        mapirow = dict((p.ulPropTag, p.Value) for p in row)
        return Rule(mapirow, table)

    def dumps(self):
        """Serialize folder contents."""
        data = {}

        data['settings'] = self.settings_dumps()
        data['items'] = [item.dumps() for item in self]

        return _utils.pickle_dumps(data)

    def loads(self, data):
        """Deserialize folder contents

        :param data: Serialized data
        """
        data = _utils.pickle_loads(data)

        self.settings_loads(data['settings'])

        for data in data['items']:
            self.create_item(loads=data)

    def settings_dumps(self):
        """Serialize folder settings (rules, permissions)."""
        data = {}

        data['rules'] = self.rules_dumps()
        data['permissions'] = self.permissions_dumps()

        return _utils.pickle_dumps(data)

    def settings_loads(self, data):
        """Deserialize folder settings (rules, permissions).

        :param data: Serialized data
        """
        data = _utils.pickle_loads(data)

        self.rules_loads(data['rules'])
        self.permissions_loads(data['permissions'])

    def rules_dumps(self, stats=None):
        """Serialize folder rules."""
        server = self.server
        log = server.log

        ruledata = None
        with _log.log_exc(log, stats):
            try:
                ruledata = self.prop(PR_RULES_DATA).value
            except (MAPIErrorNotFound, NotFoundError):
                pass
            else:
                try:
                    etxml = ElementTree.fromstring(ruledata)
                except ElementTree.ParseError:
                    log.warning("ElementTree.fromstring(ruledata): folder %s", self.name, exc_info=True)
                    return None
                for actions in etxml.findall('./item/item/actions'):
                    for movecopy in actions.findall('.//moveCopy'):
                        try:
                            s = movecopy.findall('store')[0]
                            store = server.mapisession.OpenMsgStore(0, _unbase64(s.text), None, 0)
                            entryid = HrGetOneProp(store, PR_STORE_ENTRYID).Value
                            store = server.store(entryid=_benc(entryid))
                            if store.public:
                                s.text = 'public'
                            else:
                                s.text = store.user.name if store != self.store else ''
                            f = movecopy.findall('folder')[0]
                            path = store.folder(
                                entryid=_benc(_unbase64(f.text))).path
                            f.text = path
                        except Exception:
                            log.warning("rules_dumps: could not resolve rule target", exc_info=True)
                try:
                    ruledata = ElementTree.tostring(etxml)
                except ElementTree.ParseError as e:
                    log.warning("Unable to backup rules for folder %s : %s", self.name,  str(e))
                    return None

        return _utils.pickle_dumps({
            b'data': ruledata
        })

    def rules_loads(self, data, stats=None):
        """Deserialize folder rules.

        :param data: Serialized data
        """
        server = self.server
        log = server.log

        with _log.log_exc(log, stats):
            data = _utils.pickle_loads(data)
            if isinstance(data, dict):
                data = data[b'data']
            if data:
                etxml = ElementTree.fromstring(data)
                for actions in etxml.findall('./item/item/actions'):
                    for movecopy in actions.findall('.//moveCopy'):
                        try:
                            s = movecopy.findall('store')[0]
                            if s.text == 'public':
                                store = server.public_store
                            else:
                                store = server.user(s.text).store \
                                    if s.text else self.store
                            s.text = _base64(_bdec(store.entryid))
                            f = movecopy.findall('folder')[0]
                            f.text = \
                                _base64(_bdec(store.folder(f.text).entryid))
                        except Exception as e:
                            log.warning("rules_loads: could not resolve rule target: %s", str(e), exc_info=True)
                etxml = ElementTree.tostring(etxml)
                self[PR_RULES_DATA] = etxml

    # TODO associated, PR_CONTAINER_CONTENTS?
    def table(self, name, restriction=None, order=None, columns=None):
        return Table(
            self.server,
            self.mapiobj,
            self.mapiobj.OpenProperty(name, IID_IMAPITable, MAPI_UNICODE, 0),
            name,
            restriction=restriction,
            order=order,
            columns=columns,
        )

    def tables(self): # TODO associated, rules
        yield self.table(PR_CONTAINER_CONTENTS)
        yield self.table(PR_FOLDER_ASSOCIATED_CONTENTS)
        yield self.table(PR_CONTAINER_HIERARCHY)

    @property
    def state(self):
        """Folder state (for use with ICS synchronization)."""
        return _ics.state(self.mapiobj, self._content_flag == MAPI_ASSOCIATED)

    def sync(self, importer, state=None, log=None, max_changes=None,
            associated=False, window=None, begin=None, end=None, stats=None):
        """Perform synchronization against folder

        :param importer: importer instance with callbacks to process changes
        :param state: start from this state; if not given sync from scratch
        :log: logger instance to receive important warnings/errors
        """
        if state is None:
            state = _benc(8 * b'\0')
        importer.store = self.store
        return _ics.sync(self.store.server, self.mapiobj, importer, state,
            max_changes, associated, window=window, begin=begin, end=end,
            stats=stats)

    def sync_hierarchy(self, importer, state=None, stats=None):
        """Perform hierarchy synchronization against folder. In other words,
        receive changes to the folder and its subfolders (recursively).

        :param importer: importer instance with callbacks to process changes
        :param state: start from this state; if not given sync from scratch
        """
        if state is None:
            state = _benc(8 * b'\0')
        importer.store = self.store
        return _ics.sync_hierarchy(self.store.server, self.mapiobj, importer,
            state, stats)

    # TODO merge mbox/maildir functionality
    def readmbox(self, location):
        """Import all :class:`items <Item>` from mailbox (python mailbox
        module, mbox variant).

        :param location: mailbox location
        """
        for message in mailbox.mbox(location):
            _item.Item(self, eml=message.as_bytes(unixfrom=True), create=True)

    def mbox(self, location):
        """Export all :class:`items <Item>` to mailbox (using python mailbox
        module, mbox variant).

        :param location: mailbox location
        """
        mboxfile = mailbox.mbox(location)
        mboxfile.lock()
        for item in self.items():
            mboxfile.add(item.eml())
        mboxfile.unlock()

    def read_maildir(self, location):
        """Import all :class:`items <Item>` from mailbox (python mailbox
        module, maildir variant).

        :param location: mailbox location
        """
        for message in mailbox.MH(location):
            _item.Item(self, eml=message.as_bytes(unixfrom=True), create=True)

    def maildir(self, location='.'):
        """Export all :class:`items <Item>` to mailbox (using python mailbox
        module, maildir variant).

        :param location: mailbox location
        """
        destination = mailbox.MH(location + '/' + self.name)
        destination.lock()
        for item in self.items():
            destination.add(item.eml())
        destination.unlock()

    def read_ics(self, ics):
        """Import all :class:`items <Item>` from iCal calendar

        :param ics: the iCal data
        """
        icm = icalmapi.CreateICalToMapi(self.mapiobj, self.server.ab, False)
        icm.ParseICal(ics, 'utf-8', 'UTC', self.store.user.mapiobj, 0)
        for i in range(0, icm.GetItemCount()):
            mapiobj = self.mapiobj.CreateMessage(None, 0)
            icm.GetItem(i, 0, mapiobj)
            _utils._save(mapiobj)

    def ics(self, charset="UTF-8"):
        """Export all calendar items in the folder to iCal data"""
        mic = icalmapi.CreateMapiToICal(self.server.ab, charset)
        for item in self.items():
            if item.message_class.startswith('IPM.Appointment'):
                mic.AddMessage(item.mapiobj, "", 0)
        data = mic.Finalize(0)[1]
        return data

    @property
    def associated(self):
        """Folder containing hidden items."""
        return Folder(self.store, self.entryid, associated=True)

    @property
    def deleted(self):
        """Folder containing soft-deleted items."""
        return Folder(self.store, self.entryid, deleted=True)

    def permissions(self):
        """Return all :class:`permissions <Permission>` set for this folder."""
        return _utils.permissions(self)

    def permission(self, member, create=False):
        """Return :class:`permission <Permission>` for user or group set for
        this folder.

        :param member: user or group
        :param create: create new permission for this folder
        """
        return _utils.permission(self, member, create)

    def rights(self, member):
        """Determine rights on this folder for a given :class:`user <User>`
        or :class:`group <Group>`.

        :param member: User or group
        """
        # TODO admin-over-user, Store.rights (inheritance)
        if member == self.store.user:
            return NAME_RIGHT.keys()
        parent = self
        feids = set() # avoid loops
        while parent and parent.entryid not in feids:
            try:
                return parent.permission(member).rights
            except NotFoundError:
                if isinstance(member, _user.User):
                    for group in member.groups():
                        try:
                            return parent.permission(group).rights
                        except NotFoundError:
                            pass
                    # TODO company
            feids.add(parent.entryid)
            parent = parent.parent
        return []

    def search(self, text, recurse=False):
        searchfolder = self.store.create_searchfolder()
        searchfolder.search_start(self, text, recurse)
        searchfolder.search_wait()
        for item in searchfolder:
            yield item
        self.store.findroot.mapiobj.DeleteFolder(
            _bdec(searchfolder.entryid), 0, None, 0) # TODO store.findroot

    def search_start(self, folders, text, recurse=False):
        # specific restriction format, needed to reach indexer
        restriction = SOrRestriction([
            SContentRestriction(FL_SUBSTRING | FL_IGNORECASE,
                PR_SUBJECT_W,
                SPropValue(PR_SUBJECT_W, str(text))),
            SContentRestriction(FL_SUBSTRING | FL_IGNORECASE,
                PR_BODY_W,
                SPropValue(PR_BODY_W, str(text))),
            SContentRestriction(FL_SUBSTRING | FL_IGNORECASE,
                PR_DISPLAY_TO_W,
                SPropValue(PR_DISPLAY_TO_W, str(text))),
            SContentRestriction(FL_SUBSTRING | FL_IGNORECASE,
                PR_DISPLAY_NAME_W,
                SPropValue(PR_DISPLAY_NAME_W, str(text))),
            # TODO add all default fields..
            # BUT perform full-text search by default!
        ])
        if isinstance(folders, Folder):
            folders = [folders]

        search_flags = 0
        if recurse:
            search_flags = SEARCH_RECURSIVE
        self.mapiobj.SetSearchCriteria(restriction,
            [_bdec(f.entryid) for f in folders], search_flags)

    def search_wait(self):
        while True:
            state = self.mapiobj.GetSearchCriteria(0)[-1]
            if not state & SEARCH_REBUILD:
                break

    @property
    def archive_folder(self):
        """Archive :class:`Folder`."""
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_STORE_ENTRYIDS = CHANGE_PROP_TYPE(ids[0], PT_MV_BINARY)
        PROP_ITEM_ENTRYIDS = CHANGE_PROP_TYPE(ids[1], PT_MV_BINARY)

        try:
            # support for multiple archives was a mistake, and is not and
            # _should not_ be used. so we just pick nr 0.
            arch_storeid = \
                HrGetOneProp(self.mapiobj, PROP_STORE_ENTRYIDS).Value[0]
            arch_folderid = \
                HrGetOneProp(self.mapiobj, PROP_ITEM_ENTRYIDS).Value[0]
        except MAPIErrorNotFound:
            return

        archive_store = self.server._store2(arch_storeid)
        return _store.Store(mapiobj=archive_store, server=self.server).folder(
            entryid=_benc(arch_folderid))

    @property
    def primary_store(self):
        """Primary :class:`store <Store>` (for archive folders)."""
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_REF_STORE_ENTRYID = CHANGE_PROP_TYPE(ids[3], PT_BINARY)
        try:
            entryid = HrGetOneProp(self.mapiobj, PROP_REF_STORE_ENTRYID).Value
        except MAPIErrorNotFound:
            return

        return _store.Store(entryid=_benc(entryid), server=self.server)

    @property
    def primary_folder(self):
        """Primary :class:`folder <Folder>` (for archive folders)."""
        # TODO merge namedprops stuff
        ids = self.mapiobj.GetIDsFromNames(NAMED_PROPS_ARCHIVER, MAPI_CREATE)
        PROP_REF_ITEM_ENTRYID = CHANGE_PROP_TYPE(ids[4], PT_BINARY)
        try:
            entryid = HrGetOneProp(self.mapiobj, PROP_REF_ITEM_ENTRYID).Value
        except MAPIErrorNotFound:
            return

        if self.primary_store:
            try:
                return self.primary_store.folder(entryid=_benc(entryid))
            except NotFoundError:
                pass

    @property
    def created(self):
        """Folder creation date."""
        return self._get_fast(PR_CREATION_TIME)

    @property
    def last_modified(self):
        """Folder last modification date."""
        return self._get_fast(PR_LAST_MODIFICATION_TIME)

    def subscribe(self, sink, **kwargs): # TODO document args
        """Subscribe to folder notifications

        :param sink: Sink instance with callbacks to process notifications
        :param object_types: Tracked objects (*item*, *folder*)
        :param folder_types: Tracked folders (*mail*, *contacts*, *calendar*)
        :param event_types: Event types (*created*, *updated*, *deleted*)
        """
        _notification.subscribe(self.store, self, sink, **kwargs)

    def unsubscribe(self, sink):
        """Unsubscribe from folder notifications

        :param sink: Previously subscribed sink instance
        """
        _notification.unsubscribe(self.store, sink)

    def event(self, eventid):
        eventid = _bdec(eventid)
        leid = _utils.unpack_short(eventid, 1)
        entryid = _benc(eventid[3:3 + leid])
        if not entryid:
            raise ArgumentError("invalid eventid")
        item = self.item(entryid)
        if eventid[0] == 0:
            return item
        else:
            return item.occurrence(_benc(eventid[1:]))

    @property
    def type_(self):
        """Folder type (mail, contacts, calendar, notes, tasks or just folder)."""
        if self.container_class in (None, 'IPF.Note'):
            return 'mail'
        elif self.container_class == 'IPF.Contact':
            return 'contacts'
        elif self.container_class == 'IPF.Appointment':
            return 'calendar'
        elif self.container_class == 'IPF.StickyNote':
            return 'notes'
        elif self.container_class == 'IPF.Task':
            return 'tasks'
        else:
            return 'folder'

    def permissions_dumps(self, **kwargs):
        """Serialize permissions."""
        return _permissions_dumps(self, **kwargs)

    def permissions_loads(self, data, **kwargs):
        """Deserialize permissions.

        :param data: Serialized data
        """
        return _permissions_loads(self, data, **kwargs)

    def __eq__(self, f): # TODO check same store?
        if isinstance(f, Folder):
            return self._entryid == f._entryid
        return False

    def __ne__(self, f):
        return not self == f

    def __iter__(self):
        self._iter = self.items()
        return self

    def __next__(self):
        if not self._iter:
            self._iter = self.items()
        return next(self._iter)

    def next(self):
        return self.__next__()

    def __unicode__(self): # TODO associated?
        try: # folder deleted via ICS
            name = self.name
        except AttributeError:
            name = ''

        return 'Folder(%s)' % name



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)
"""

import binascii
import datetime
import os
try:
    import cPickle as pickle
except ImportError:
    import _pickle as pickle
import struct
import time

from MAPI import (
    WrapCompressedRTFStream, PT_UNICODE, ROW_ADD, MAPI_MODIFY,
    KEEP_OPEN_READWRITE,
)
from MAPI.Defs import (
    PROP_TYPE
)
from MAPI.Tags import (
    PR_RTF_COMPRESSED, PR_ACL_TABLE, PR_MEMBER_ENTRYID, PR_MEMBER_RIGHTS
)
from MAPI.Tags import (
    IID_IStream, IID_IECMessageRaw, IID_IExchangeModifyTable
)
from MAPI.Time import (
    FileTime, NANOSECS_BETWEEN_EPOCH,
)
from MAPI.Struct import (
    MAPIErrorNotFound, MAPIErrorInterfaceNotSupported, SPropValue, ROWENTRY,
    MAPIErrorNoAccess, MAPIErrorDiskError
)

TESTING = False
if os.getenv('PYKO_TESTING'): # env variable used in testset
    TESTING = True

MAX_SAVE_RETRIES = int(os.getenv('PYKO_MAPI_SAVE_MAX_RETRIES', 3))

from .compat import bdec as _bdec
from .errors import Error, NotFoundError, ArgumentError

from . import table as _table
from . import permission as _permission
from . import user as _user
from . import group as _group

def pickle_loads(s):
    return pickle.loads(s, encoding='bytes')

def pickle_dumps(s):
    return pickle.dumps(s, protocol=2)

def stream(mapiobj, proptag):
    stream = mapiobj.OpenProperty(proptag, IID_IStream, 0, 0)

    if proptag == PR_RTF_COMPRESSED:
        stream = WrapCompressedRTFStream(stream, 0)

    block_size = 0x100000 # 1MB

    data = []
    while True:
        temp = stream.Read(block_size)
        data.append(temp)

        if len(temp) < block_size:
            break

    data = b''.join(data)

    if PROP_TYPE(proptag) == PT_UNICODE:
        data = data.decode('utf-32le') # under windows them be utf-16le?

    return data

# avoid underwater action for archived items
def _openentry_helper(mapistore, entryid, flags):
    try:
        return mapistore.OpenEntry(entryid, IID_IECMessageRaw, flags)
    except MAPIErrorInterfaceNotSupported:
        return mapistore.OpenEntry(entryid, None, flags)

# try to open read/write, falling back to read-only
def openentry_raw(mapistore, entryid, flags):
    try:
        return _openentry_helper(mapistore, entryid, flags | MAPI_MODIFY)
    except MAPIErrorNoAccess:
        return _openentry_helper(mapistore, entryid, flags)

def unpack_short(s, pos):
    return struct.unpack_from('<H', s, pos)[0]

def unpack_long(s, pos):
    return struct.unpack_from('<L', s, pos)[0]

def pack_short(i):
    return struct.pack('<H', i)

def pack_long(i):
    return struct.pack('<L', i)

def rectime_to_unixtime(t):
    return (t - 194074560) * 60

def unixtime_to_rectime(t):
    return int(t / 60) + 194074560

def permissions(obj):
        try:
            acl_table = obj.mapiobj.OpenProperty(
                PR_ACL_TABLE, IID_IExchangeModifyTable, 0, 0)
        except MAPIErrorNotFound:
            return
        table = _table.Table(
            obj.server,
            obj.mapiobj,
            acl_table.GetTable(0),
            PR_ACL_TABLE,
        )
        for row in table.dict_rows():
            yield _permission.Permission(acl_table, row, obj.server)

def permission(obj, member, create):
        for permission in obj.permissions():
            if permission.member == member:
                return permission
        if create:
            acl_table = obj.mapiobj.OpenProperty(
                PR_ACL_TABLE, IID_IExchangeModifyTable, 0, 0)

            if isinstance(member, _user.User): # TODO *.id_ or something..?
                memberid = member.userid
            elif isinstance(member, _group.Group):
                memberid = member.groupid
            else:
                memberid = member.companyid

            props = [
                SPropValue(PR_MEMBER_ENTRYID, _bdec(memberid)),
                SPropValue(PR_MEMBER_RIGHTS, 0)
            ]
            acl_table.ModifyTable(0, [ROWENTRY(ROW_ADD, props)])
            return obj.permission(member)
        else:
            raise NotFoundError("no permission entry for '%s'" % member.name)

def bytes_to_human(b):
    suffixes = ['b', 'kb', 'mb', 'gb', 'tb', 'pb']
    if b == 0: return '0 b'
    i = 0
    len_suffixes = len(suffixes) - 1
    while b >= 1024 and i < len_suffixes:
        b //= 1024
        i += 1
    f = ('%.2f' % b).rstrip('0').rstrip('.')
    return '%s %s' % (f, suffixes[i])

def human_to_bytes(s):
    """
    Author: Giampaolo Rodola' <g.rodola [AT] gmail [DOT] com>
    License: MIT
    """
    s = s.lower()
    init = s
    num = ""
    while s and s[0:1].isdigit() or s[0:1] == '.':
        num += s[0]
        s = s[1:]
    try:
        num = float(num)
    except ValueError:
        raise ArgumentError('invalid size: %r' % init)
    letter = s.strip()
    for sset in [
        ('b', 'k', 'm', 'g', 't', 'p', 'e', 'z', 'y'),
        ('b', 'kb', 'mb', 'gb', 'tb', 'pb', 'eb', 'zb', 'yb'),
        ('b', 'kib', 'mib', 'gib', 'tib', 'pib', 'eib', 'zib', 'yib')
    ]:
        if letter in sset:
            break
    else:
        raise ArgumentError('invalid size: %r' % init)
    prefix = {sset[0]: 1}
    for i, s in enumerate(sset[1:]):
        prefix[s] = 1 << (i + 1) * 10
    return int(num * prefix[letter])

def arg_objects(arg, supported_classes, method_name):
    if isinstance(arg, supported_classes):
        objects = [arg]
    else:
        try:
            objects = list(arg)
        except TypeError:
            raise ArgumentError('invalid argument to %s' % method_name)

    if [o for o in objects if not isinstance(o, supported_classes)]:
        raise ArgumentError('invalid argument to %s' % method_name)
    return objects

def _bdec_eid(entryid):
    try:
        return _bdec(entryid)
    except (TypeError, AttributeError, binascii.Error):
        raise ArgumentError("invalid entryid: %r" % entryid) from None

def _save(mapiobj):
    # retry on deadlock or other temporary issue
    t = 0.1
    retry = 0
    while True:
        try:
            if TESTING and os.getenv('PYKO_TEST_DISK_ERROR'): # test coverage
                raise MAPIErrorDiskError()
            mapiobj.SaveChanges(KEEP_OPEN_READWRITE)
            break
        except MAPIErrorDiskError:
            if retry >= MAX_SAVE_RETRIES:
                raise Error('could not save object')
            else:
                retry += 1
                time.sleep(t)

def datetime_to_filetime(d):
    return FileTime(int(datetime.datetime.timestamp(d)) * 10000000 + NANOSECS_BETWEEN_EPOCH)



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2016 - Kopano and its licensors (see LICENSE file for details)
"""

from MAPI.Struct import MAPIErrorNotFound
from MAPI.Util import TestRestriction


class Restriction(object):
    """Restriction class"""

    def __init__(self, mapiobj=None):
        self.mapiobj = mapiobj

    def match(self, item):
        try:
            TestRestriction(self.mapiobj, item.mapiobj)
            return True
        except MAPIErrorNotFound:
            return False

    def __unicode__(self):
        return u'Restriction()'

    def __repr__(self):
        return self.__unicode__()



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-only
from .version import __version__

import collections
from contextlib import closing
import fcntl
import os.path

from multiprocessing import Queue, Value
import time
import sys

import bsddb3 as bsddb
from queue import Empty

from kopano_search import plaintext
import kopano
from kopano import log_exc, Config
sys.path.insert(0, os.path.dirname(__file__)) # XXX for __import__ to work

"""
kopano-search v3 - a process-based indexer and query handler built on python-kopano

initial indexing is performed in parallel using N instances of class IndexWorker, fed with a queue containing folder-ids.

incremental indexing is later performed in the same fashion. when there is no pending work, we
check if there are reindex requests and handle one of these (again parallellized). so incremental syncing is currently paused
during reindexing.

search queries from outlook/webapp are dealt with by a single instance of class SearchWorker.

since ICS does not know for deletion changes which store they belong to, we remember ourselves using a berkeleydb file ("serverguid_mapping").

we also maintain folder and server ICS states in a berkeleydb file, for example so we can resume initial indexing ("serverguid_state").
after initial indexing folder states are not updated anymore.

the used search engine is pluggable, but by default we use xapian (plugin_xapian.py).

a tricky bit is that outlook/exchange perform 'prefix' searches by default and we want to be compatible with this, so terms get an
implicit '*' at the end. not every search engine may perform well for this, but we could make this configurable perhaps.

"""

CONFIG = {
    'index_attachments': Config.boolean(default=False),
    'index_attachment_extension_filter': Config.ignore(),
    'index_attachment_mime_filter': Config.ignore(),
    'index_attachment_max_size': Config.size(default=2**24),
    'index_attachment_parser': Config.ignore(),
    'index_attachment_parser_max_memory': Config.ignore(),
    'index_attachment_parser_max_cputime': Config.ignore(),
    # 0x007D: PR_TRANSPORT_MESSAGE_HEADERS
    # 0x0064: PR_SENT_REPRESENTING_ADDRTYPE
    # 0x0C1E: PR_SENDER_ADDRTYPE
    # 0x0075: PR_RECEIVED_BY_ADDRTYPE
    # 0x678E: PR_EC_IMAP_BODY
    # 0x678F: PR_EC_IMAP_BODYSTRUCTURE
    # 0x001A: PR_MESSAGE_CLASS # XXX add to cfg
    'index_exclude_properties': Config.integer(multiple=True, base=16, default=[0x007D, 0x0064, 0x0C1E, 0x0075, 0x678E, 0x678F, 0x001A]),
    'index_path': Config.string(default='/var/db/kopano/search/'),
    'index_processes': Config.integer(default=1),
    'limit_results': Config.integer(default=1000),
    'run_as_user': Config.string(default="_kopano"),
    'run_as_group': Config.string(default="_kopano"),
    'search_engine': Config.string(default='xapian'),
    'suggestions': Config.boolean(default=True),
    'index_junk': Config.boolean(default=True),
    'index_drafts': Config.boolean(default=True),
    'server_bind_name': Config.string(default='file:///var/run/kopano/search.sock'),
    'ssl_private_key_file': Config.path(default=None, check=False), # XXX don't check when default=None?
    'ssl_certificate_file': Config.path(default=None, check=False),
    'term_cache_size': Config.size(default=64000000),
}

def db_get(db_path, key):
    """ get value from db file """
    if not isinstance(key, bytes):  # python3
        key = key.encode('ascii')
    with closing(bsddb.hashopen(db_path, 'c')) as db:
        value = db.get(key)
        if value is not None:
            return db.get(key).decode('ascii')

def db_put(db_path, key, value):
    """ store key, value in db file """
    if not isinstance(key, bytes):  # python3
        key = key.encode('ascii')
    with open(db_path+'.lock', 'w') as lockfile:
        fcntl.flock(lockfile.fileno(), fcntl.LOCK_EX)
        with closing(bsddb.hashopen(db_path, 'c')) as db:
            db[key] = value

class SearchWorker(kopano.Worker):
    """ process which handles search requests coming from outlook/webapp, according to our internal protocol """

    def main(self):
        config, plugin = self.service.config, self.service.plugin
        def response(conn, msg):
            self.log.info('Response: %s', msg)
            conn.sendall((msg + '\r\n').encode())
        s = kopano.server_socket(config['server_bind_name'], ssl_key=config['ssl_private_key_file'], ssl_cert=config['ssl_certificate_file'], log=self.log)
        while True:
            with log_exc(self.log):
                conn, _ = s.accept()
                fields_terms = []
                for data in conn.makefile():
                    data = data.strip()
                    self.log.info('Command: %s', data)
                    cmd, args = data.split()[0], data.split()[1:]
                    if cmd == 'PROPS':
                        response(conn, 'OK:'+' '.join(map(str, config['index_exclude_properties'])))
                        break
                    if cmd == 'SYNCRUN':  # wait for syncing to be up-to-date (only used in tests)
                        self.syncrun.value = time.time()
                        while self.syncrun.value:
                            time.sleep(1)
                        response(conn, 'OK:')
                        break
                    elif cmd == 'SCOPE':
                        server_guid, store_guid, folder_ids = args[0], args[1], args[2:]
                        response(conn, 'OK:')
                    elif cmd == 'FIND':
                        pos = data.find(':')
                        fields = [int(x) for x in data[:pos].split()[1:]]
                        orig = data[pos+1:].lower()
                        # Limit number of terms (32) so people do not
                        # inadvertently DoS it if they paste prose.
                        terms = plugin.extract_terms(orig)[:32]
                        if fields and terms:
                            fields_terms.append((fields, terms))
                        response(conn, 'OK:')
                    elif cmd == 'SUGGEST':
                        suggestion = u''
                        if config['suggestions'] and len(fields_terms) == 1:
                            for fields, terms in fields_terms:
                                suggestion = plugin.suggest(server_guid, store_guid, terms, orig, self.log)
                                if suggestion == orig:
                                    suggestion = u''
                        response(conn, 'OK: '+suggestion)
                    elif cmd == 'QUERY':
                        t0 = time.time()
                        restrictions = []
                        if folder_ids:
                            restrictions.append('('+' OR '.join(['folderid:%s' % f for f in folder_ids])+')')
                        for fields, terms in fields_terms:
                            if fields:
                                restrictions.append('('+' OR '.join('('+' AND '.join('mapi%d:%s*' % (f, term) for term in terms)+')' for f in fields)+')')
                            else:
                                restrictions.append('('+' AND '.join('%s*' % term for term in terms)+')')
                        query = ' AND '.join(restrictions)  # plugin doesn't have to use this relatively standard query format
                        docids = plugin.search(server_guid, store_guid, folder_ids, fields_terms, query, config['limit_results'], self.log)
                        response(conn, 'OK: '+' '.join(map(str, docids)))
                        self.log.info('found %d results in %.2f seconds', len(docids), time.time()-t0)
                        break
                    elif cmd == 'REINDEX':
                        self.reindex_queue.put(args[0])
                        response(conn, 'OK:')
                        self.log.info("queued store %s for reindexing", args[0])
                        break
                    else:
                        self.log.error("unknown command: %s", cmd)
                        break
                conn.close()

class IndexWorker(kopano.Worker):
    """ process which gets folders from input queue and indexes them, putting the nr of changes in output queue """

    def main(self):
        config, server, plugin = self.service.config, self.server, self.service.plugin
        state_db = os.path.join(config['index_path'], server.guid+'_state')
        while True:
            changes = 0
            with log_exc(self.log):
                (_, storeguid, folderid, reindex) = self.iqueue.get()
                store = server.store(storeguid)
                folder = kopano.Folder(store, folderid)
                path = folder.path
                if path and \
                   (folder != store.outbox) and \
                   (folder != store.junk or config['index_junk']) and \
                   (folder != store.drafts or config['index_drafts']):
                    suggestions = config['suggestions'] and folder != store.junk
                    self.log.info('syncing folder: "%s" "%s"', store.name, path)
                    importer = FolderImporter(server.guid, config, plugin, suggestions, self.log)
                    state = db_get(state_db, folder.entryid) if not reindex else None
                    if state:
                        self.log.info('found previous folder sync state: %s', state)
                    t0 = time.time()
                    new_state = folder.sync(importer, state, log=self.log)
                    if new_state != state:
                        plugin.commit(suggestions)
                        db_put(state_db, folder.entryid, new_state)
                        self.log.info('saved folder sync state: %s', new_state)
                        changes = importer.changes + importer.deletes
                        self.log.info('syncing folder "%s" took %.2f seconds (%d changes, %d attachments)', path, time.time()-t0, changes, importer.attachments)
            self.oqueue.put(changes)

class FolderImporter:
    """ tracks changes for a given folder """

    def __init__(self, *args):
        self.serverid, self.config, self.plugin, self.suggestions, self.log = args
        self.changes = self.deletes = self.attachments = 0
        self.mapping_db = os.path.join(self.config['index_path'], self.serverid+'_mapping')
        self.excludes = set(self.config['index_exclude_properties']+[0x1000, 0x1009, 0x1013, 0x678C, 0x6791]) # PR_BODY, PR_RTF_COMPRESSED, PR_HTML, PR_EC_IMAP_EMAIL, PR_EC_BODY_FILTERED
        self.term_cache_size = 0

    def update(self, item, flags):
        """ called for a new or changed item; get mapi properties, attachments and pass to indexing plugin """

        with log_exc(self.log):
            self.changes += 1
            storeid, folderid, entryid, sourcekey, docid = item.storeid, item.folder.hierarchyid, item.entryid, item.sourcekey, item.docid
            self.log.debug('store %s, folder %d: new/updated document with entryid %s, sourcekey %s, docid %d', storeid, folderid, entryid, sourcekey, docid)

            doc = collections.defaultdict(str)
            doc.update({'serverid': self.serverid, 'storeid': storeid, 'folderid': folderid, 'docid': docid, 'sourcekey': item.sourcekey})
            attach_text = []

            for subitem in [item] + list(item.items()):
                for prop in subitem.props():
                    if prop.id_ not in self.excludes:
                        if isinstance(prop.value, str):
                            if prop.value:
                                doc['mapi%d' % prop.id_] += u' ' + prop.value
                        elif isinstance(prop.value, list):
                            doc['mapi%d' % prop.id_] += u' ' + u' '.join(x for x in prop.value if isinstance(x, str))

                for attachment in subitem.attachments():
                    if self.config['index_attachments'] and \
                       0 < len(attachment) < self.config['index_attachment_max_size'] and \
                       attachment.filename != 'inline.txt':  # XXX inline attachment check
                        self.log.debug('indexing attachment (filename=%s, size=%d, mimetag=%s)',
                            attachment.filename, len(attachment), attachment.mimetype)
                        self.attachments += 1
                        attach_text.append(plaintext.get(attachment, mimetype=attachment.mimetype, log=self.log))
                    attach_text.append(attachment.filename or u'')

                doc['mapi4096'] += u' ' + subitem.text + u' ' + u' '.join(attach_text)  # PR_BODY
                doc['mapi3098'] += u' ' + u' '.join([subitem.sender.name, subitem.sender.email, subitem.from_.name, subitem.from_.email])  # PR_SENDER_NAME
                doc['mapi3588'] += u' ' + u' '.join([a.name + u' ' + a.email for a in subitem.to])  # PR_DISPLAY_TO
                doc['mapi3587'] += u' ' + u' '.join([a.name + u' ' + a.email for a in subitem.cc])  # PR_DISPLAY_CC
                doc['mapi3586'] += u' ' + u' '.join([a.name + u' ' + a.email for a in subitem.bcc])  # PR_DISPLAY_BCC

            doc['data'] = 'subject: %s\n' % item.subject
            db_put(self.mapping_db, item.sourcekey, '%s %s' % (storeid, item.folder.entryid))  # ICS doesn't remember which store a change belongs to..
            self.plugin.update(doc)
            self.term_cache_size += sum(len(v) for k, v in doc.items() if k.startswith('mapi'))
            if (8*self.term_cache_size) > self.config['term_cache_size']:
                self.plugin.commit(self.suggestions)
                self.term_cache_size = 0

    def delete(self, item, flags):
        """ for a deleted item, determine store and ask indexing plugin to delete """

        with log_exc(self.log):
            self.deletes += 1
            ids = db_get(self.mapping_db, item.sourcekey)
            if ids:  # when a 'new' item is deleted right away (spooler?), the 'update' function may not have been called
                storeid, folderid = ids.split()
                doc = {'serverid': self.serverid, 'storeid': storeid, 'sourcekey': item.sourcekey}
                self.log.debug('store %s: deleted document with sourcekey %s', doc['storeid'], item.sourcekey)
                self.plugin.delete(doc)

class ServerImporter:
    """ tracks changes for a server node; queues encountered folders for updating """  # XXX improve ICS to track changed folders?

    def __init__(self, serverid, config, iqueue, log):
        self.mapping_db = os.path.join(config['index_path'], serverid+'_mapping')
        self.iqueue = iqueue
        self.queued = set() # sync each folder at most once

    def update(self, item, flags):
        self.queue((0, item.storeid, item.folder.entryid))

    def delete(self, item, flags):
        ids = db_get(self.mapping_db, item.sourcekey)
        if ids:
            self.queue((0,) + tuple(ids.split()))

    def queue(self, folder):
        if folder not in self.queued:
            self.iqueue.put(folder+(False,))
            self.queued.add(folder)

class Service(kopano.Service):
    """ main search process """

    def main(self):
        """ start initial syncing if no state found. then start query process and switch to incremental syncing """

        self.reindex_queue = Queue()
        index_path = self.config['index_path']
        os.umask(0o77)
        if not os.path.exists(index_path):
            try:
                os.makedirs(index_path)
            except PermissionError:
                self.log.error("Unable to create directory '%s': permission denied", index_path)
                sys.exit(1)
        self.state_db = os.path.join(index_path, self.server.guid+'_state')
        self.plugin = __import__('plugin_%s' % self.config['search_engine']).Plugin(index_path, self.log)
        self.iqueue, self.oqueue = Queue(), Queue()
        self.index_processes = self.config['index_processes']
        workers = [IndexWorker(self, 'index%d'%i, nr=i, iqueue=self.iqueue, oqueue=self.oqueue) for i in range(self.index_processes)]
        for worker in workers:
            worker.start()
        try:
            self.state = db_get(self.state_db, 'SERVER')
        except bsddb.db.DBAccessError:
            self.log.error("Cannot access '%s': permission denied", self.state_db)
            sys.exit(1)

        if self.state:
            self.log.info('found previous server sync state: %s', self.state)
        else:
            self.log.info('starting initial sync')
            new_state = self.server.state  # syncing will reach at least the current state
            self.initial_sync(self.server.stores())
            self.state = new_state
            db_put(self.state_db, 'SERVER', self.state)
            self.log.info('saved server sync state = %s', self.state)
        self.syncrun = Value('d', 0)
        SearchWorker(self, 'query', reindex_queue=self.reindex_queue, syncrun=self.syncrun).start()
        self.log.info('starting incremental sync')
        self.incremental_sync()

    def initial_sync(self, stores, reindex=False):
        """ queue all folders for given stores """

        folders = [(f.count, s.guid, f.entryid, reindex) for s in stores for f in s.folders()]
        for f in sorted(folders, reverse=True):
            self.iqueue.put(f)
        itemcount = sum(f[0] for f in folders)
        self.log.info('queued %d folders (~%d changes) for parallel indexing (%s processes)', len(folders), itemcount, self.index_processes)
        t0 = time.time()
        changes = sum([self.oqueue.get() for i in range(len(folders))])  # blocking
        self.log.info('queue processed in %.2f seconds (%d changes, ~%.2f/sec)', time.time()-t0, changes, changes/(time.time()-t0))

    def incremental_sync(self):
        """ process changes in real-time (not yet parallelized); if no pending changes handle reindex requests """

        while True:
            with log_exc(self.log):
                try:
                    storeid = self.reindex_queue.get(block=False)
                    store = self.server.store(storeid)
                    self.log.info('handling reindex request for "%s"', store.name)
                    self.plugin.reindex(self.server.guid, store.guid)
                    self.initial_sync([store], reindex=True)
                except Empty:
                    pass
                importer = ServerImporter(self.server.guid, self.config, self.iqueue, self.log)
                t0 = time.time()
                new_state = self.server.sync(importer, self.state, log=self.log)
                if new_state != self.state:
                    changes = sum([self.oqueue.get() for i in range(len(importer.queued))])  # blocking
                    for f in importer.queued:
                        self.iqueue.put(f+(False,))  # make sure folders are at least synced to new_state
                    changes += sum([self.oqueue.get() for i in range(len(importer.queued))])  # blocking
                    self.log.info('queue processed in %.2f seconds (%d changes, ~%.2f/sec)', time.time()-t0, changes, changes/(time.time()-t0))
                    self.state = new_state
                    db_put(self.state_db, 'SERVER', self.state)
                    self.log.info('saved server sync state = %s', self.state)
                if t0 > self.syncrun.value+1:
                    self.syncrun.value = 0
            time.sleep(5)

    def reindex(self):
        """ pass usernames/store-ids given on command-line to running search process """

        store_guids = []

        if not (self.options.users or self.options.stores):
            print("no user/store specified (use -u/-S)", file=sys.stderr)
            sys.exit(1)

        for username in self.options.users:
            user = self.server.get_user(username)
            if not user:
                print("no such user: %s" % username, file=sys.stderr)
                sys.exit(1)
            elif user.store:
                store_guids.append(user.store.guid)

        for store_guid in self.options.stores:
            store = self.server.get_store(store_guid)
            if not store:
                print("no such store: %s" % store_guid, file=sys.stderr)
                sys.exit(1)
            else:
                store_guids.append(store.guid)

        try:
            for store_guid in store_guids:
                with closing(kopano.client_socket(self.config['server_bind_name'], ssl_cert=self.config['ssl_certificate_file'])) as s:
                    s.sendall((u'REINDEX %s\r\n' % store_guid).encode('ascii'))
                    s.recv(1024)
        except OSError as e:
            print('could not connect to search socket (%s)' % e)
            sys.exit(1)

def main():
    parser = kopano.parser('CKQSFlVus')  # select common cmd-line options
    parser.add_option('-r', '--reindex', dest='reindex', action='store_true',help='Reindex user/store')
    options, args = parser.parse_args()
    service = Service('search', config=CONFIG, options=options)
    if options.reindex:
        service.reindex()
    else:
        service.start()

if __name__ == '__main__':
    main()  # pragma: no cover



# SPDX-License-Identifier: AGPL-3.0-only
"""
Part of the high-level python bindings for Kopano

Copyright 2005 - 2016 Zarafa and its licensors (see LICENSE file)
Copyright 2016 - 2019 Kopano and its licensors (see LICENSE file)

Parts of this file are based on code from Vinay Sajip as found at
https://gist.github.com/vsajip/820132

Copyright 2011 - Vinay Sajip. All Rights Reserved.

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Vinay Sajip
not be used in advertising or publicity pertaining to distribution
of the software without specific, written prior permission.
VINAY SAJIP DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL
VINAY SAJIP BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""

import contextlib
import logging.handlers
import os
from queue import Empty
import sys
import threading
import traceback

FMT_BASE = '[%(levelname)-8s] %(name)s - %(message)s'
FMT = '%(asctime)s.%(msecs)03d: ' + FMT_BASE

def _kopano_logger():
    log = logging.getLogger('kopano')
    log.setLevel(logging.WARNING)
    fh = logging.StreamHandler(sys.stderr)
    log.addHandler(fh)
    formatter = logging.Formatter(FMT, datefmt='%Y-%m-%dT%H:%M:%S')
    fh.setFormatter(formatter)
    return log

# general python-kopano logger, called 'kopano'
# by default it records all logging from python-kopano, such as
# warnings about weird situations
LOG = _kopano_logger()

def _loglevel(options, config):
    log_level = None
    if options and getattr(options, 'loglevel', None):
        log_level = options.loglevel
    elif config:
        log_level = config.get('log_level')
    log_level = log_level or 'warning'
    return { # TODO NONE?
        '0': logging.NOTSET,
        '1': logging.CRITICAL,
        '2': logging.ERROR,
        '3': logging.WARNING,
        '4': logging.INFO,
        '5': logging.INFO,
        '6': logging.DEBUG,
        'DEBUG': logging.DEBUG,
        'INFO': logging.INFO,
        'WARNING': logging.WARNING,
        'ERROR': logging.ERROR,
        'CRITICAL': logging.CRITICAL,
    }[log_level.upper()]

def _logfmt(config):
    log_fmt = FMT
    if config:
        if not config.get('log_timestamp', True):
            log_fmt = FMT_BASE
    return log_fmt

def logger(service, options=None, stdout=False, config=None, name=''):
    logger = logging.getLogger(name or service)
    if logger.handlers:
        return logger
    formatter = logging.Formatter(_logfmt(config), datefmt='%Y-%m-%dT%H:%M:%S')
    log_method = 'file'
    log_file = '/var/log/kopano/%s.log' % service
    if config:
        log_method = config.get('log_method') or log_method
        log_file = config.get('log_file') or log_file
    log_level = _loglevel(options, config)
    if name:
        log_file = log_file.replace(service, name) # TODO
    fh = None
    if log_method == 'syslog' and not os.getenv("PYKO_ALLOW_SYSLOG_LOGGING", None):
        # NOTE(longsleep): Syslog logging does not work in its current
        # implementation. Disable it for now until it can be brought back in
        # way which works with multi-threading and multi-processing.
        import warnings
        warnings.warn('syslog support has been removed, falling back to stderr')
        log_method = 'file'
        log_file = '-'
    if log_method == 'file':
        if log_file == '-':
            fh = logging.StreamHandler(sys.stderr)
        else:
            fh = logging.handlers.WatchedFileHandler(log_file)
    elif log_method == 'syslog':
        fh = logging.handlers.SysLogHandler(address='/dev/log')
    if fh:
        fh.setLevel(log_level)
        fh.setFormatter(formatter)
        logger.addHandler(fh)
    if not getattr(options, 'service', True):
        ch = logging.StreamHandler(sys.stdout)
        ch.setLevel(log_level)
        ch.setFormatter(formatter)
        if stdout or ((log_method, log_file) != ('file', '-')):
            logger.addHandler(ch)
    logger.setLevel(log_level)
    return logger

# it logs errors, that's all you need to know :-)
@contextlib.contextmanager
def log_exc(log, stats=None):
    """
Context-manager to log any exception in sub-block to given logger instance

:param log: logger instance

Example usage::

    with log_exc(log):
        .. # any exception will be logged when exiting sub-block

"""
    try: yield
    except Exception:
        log.error(traceback.format_exc())
        if stats is not None:
            if 'errors' not in stats:
                stats['errors'] = 0
            stats['errors'] += 1

# TODO use python3 builtins if available (log.handlers.*?)

# log-to-queue handler copied from Vinay Sajip
class QueueHandler(logging.Handler): # pragma: no cover
    def __init__(self, queue):
        logging.Handler.__init__(self)
        self.queue = queue

    def enqueue(self, record):
        self.queue.put(record, True, 0.5)

    def prepare(self, record):
        self.format(record)
        record.msg, record.args, record.exc_info = record.message, None, None
        return record

    def emit(self, record):
        try:
            self.enqueue(self.prepare(record))
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)

# log-to-queue listener copied from Vinay Sajip
class QueueListener(object): # pragma: no cover
    def __init__(self, queue, *handlers):
        self.queue = queue
        self.handlers = handlers
        self._stop = threading.Event()
        self._thread = None

    def start(self):
        self._thread = t = threading.Thread(target=self._monitor)
        t.setDaemon(True)
        t.start()

    def handle(self, record):
        for handler in self.handlers:
            handler.handle(record)

    def _monitor(self):
        q = self.queue
        has_task_done = hasattr(q, 'task_done')

        while True:
            try:
                record = self.queue.get(True, 1) # block, timeout 1 sec
                self.handle(record)
                if has_task_done:
                    q.task_done()
                if self._stop.isSet():
                    break
            except Empty:
                if self._stop.isSet():
                    break
            except (EOFError, ValueError):
                if self._stop.isSet():
                    break
                raise

    def stop(self):
        self._stop.set()
        self._thread.join()
        self._thread = None



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-only
from .version import __version__

import codecs
from contextlib import closing
import csv
import datetime
import fcntl
from multiprocessing import Queue
import os.path
import re
import shutil
import sys
import time
import zlib
try:
    import cPickle as pickle
except ImportError:
    import _pickle as pickle
import bsddb3 as bsddb

from MAPI import (
    PT_UNICODE, PT_ERROR, KEEP_OPEN_READWRITE,
)

from MAPI.Defs import (
    PROP_TYPE, CHANGE_PROP_TYPE,
)

from MAPI.Struct import (
    MAPIErrorNotFound, SPropValue,
)

from MAPI.Tags import (
    PR_EC_BACKUP_SOURCE_KEY, PR_EC_OUTOFOFFICE, PR_ENTRYID, PR_DISPLAY_NAME_W,
    PR_EC_OUTOFOFFICE_SUBJECT, PR_EC_OUTOFOFFICE_MSG, PR_EC_OUTOFOFFICE_FROM,
    PR_EC_OUTOFOFFICE_UNTIL, PR_CONTAINER_CLASS_W, PR_SOURCE_KEY,
    PT_TSTRING, PR_EC_WEBACCESS_SETTINGS_W, PR_EC_RECIPIENT_HISTORY_W,
    PR_EC_WEBACCESS_SETTINGS_JSON_W, PR_EC_RECIPIENT_HISTORY_JSON_W,
    PR_EC_WEBAPP_PERSISTENT_SETTINGS_JSON_W, PR_STORE_ENTRYID
)

SETTINGS_PROPTAGS = ( # TODO long-term move to store.settings_dumps()
    PR_EC_WEBACCESS_SETTINGS_W, PR_EC_RECIPIENT_HISTORY_W,
    PR_EC_WEBACCESS_SETTINGS_JSON_W, PR_EC_RECIPIENT_HISTORY_JSON_W,
    PR_EC_WEBAPP_PERSISTENT_SETTINGS_JSON_W,
    PR_EC_OUTOFOFFICE_SUBJECT, PR_EC_OUTOFOFFICE_MSG,
    PR_EC_OUTOFOFFICE, PR_EC_OUTOFOFFICE_FROM, PR_EC_OUTOFOFFICE_UNTIL,
)

UNESCAPED_SLASH_RE = re.compile(r'(?<!\\)/')

import kopano
from kopano import log_exc, Config

"""
kopano-backup - a MAPI-level backup/restore tool built on python-kopano.

backup is done incrementally using ICS and can be parallellized over stores.

restore is not parallelized.

items are serialized and maintained in per-folder key-value stores.

metadata such as webapp settings, rules, acls and delegation permissions are also stored per-folder.

differential backups are possible via the --differential option.

basic commands (see --help for all options):

kopano-backup -u user1 -> backup (sync) data for user 'user1' in (new) directory 'user1'
kopano-backup -u user1 -f Inbox -> backup 'Inbox' folder for user 'user1' in (new) directory 'user1'

kopano-backup --restore user1 -> restore data from directory 'user1' to account called 'user1'
kopano-backup --restore -u user2 user1 -> same, but restore to account 'user2'

kopano-backup --stats user1 -> summarize contents of backup directory 'user1', in CSV format
kopano-backup --index user1 -> low-level overview of stored items, in CSV format

options can be combined when this makes sense, for example:

kopano-backup --index user1 -f Inbox/subfolder --recursive --period-begin 2014-01-01

"""

CONFIG = {
    'backup_servers': Config.string(multiple=True, default=None),
}

CACHE_SIZE = 64000000 # XXX make configurable

def pickle_dumps(s):
    return pickle.dumps(s, protocol=2)

def pickle_loads(s):
    return pickle.loads(s, encoding='bytes')

def _hex(s):
    return codecs.encode(s, 'hex').upper()

def _unhex(s):
    return codecs.decode(s, 'hex')

def fatal(s):
    sys.stderr.write(s + "\n")
    sys.exit(1)

def dbopen(path):
    return bsddb.hashopen(path, 'c')

def _copy_folder_meta(from_dir, to_dir, keep_db=False):
    if not os.path.exists(to_dir):
        os.makedirs(to_dir)

    for filename in ('acl', 'folder', 'rules', 'items', 'index', 'state', 'path'): # 'path' last to indicate completion
        if not (keep_db and filename in ('index', 'items')):
            from_path = from_dir+'/'+filename
            to_path = to_dir+'/'+filename
            if os.path.exists(from_path):
                shutil.copy(from_path, to_dir) # overwrites

def _copy_store_meta(from_dir, to_dir):
    for filename in ('delegates', 'store', 'user'):
        from_path = from_dir+'/'+filename
        to_path = to_dir+'/'+filename
        if os.path.exists(from_path):
            shutil.copy(from_path, to_dir) # overwrites

def _mark_deleted( index, fpath, timestamp, log):
    log.debug("marking deleted folder '%s'", fpath)

    with closing(dbopen(index)) as db_index:
        idx = db_index.get(b'folder')
        d = pickle_loads(idx) if idx else {}
        if not d.get(b'backup_deleted'):
            d[b'backup_deleted'] = timestamp
            db_index[b'folder'] = pickle_dumps(d)

class BackupWorker(kopano.Worker):
    """ each worker takes stores from a queue, and backs them up to disk (or syncs them),
        according to the given command-line options """

    def main(self):
        config, server, options = self.service.config, self.service.server, self.service.options
        while True:
            if self.iqueue.empty():
                continue
            stats = {'changes': 0, 'deletes': 0, 'errors': 0}
            self.service.stats = stats # XXX generalize
            with log_exc(self.log, stats):
                # get store from input queue
                (store_entryid, username, path) = self.iqueue.get()
                store = server.store(entryid=store_entryid)
                user = store.user

                # from- and to- paths for differential
                self.orig_path = path
                if options.differential:
                    path = 'differential/' + path

                # create main directory
                if not os.path.isdir(path):
                    os.makedirs(path)

                # differential: determine to- and from- backups
                if options.differential:
                    diff_ids = [int(x) for x in os.listdir(path)]
                    if diff_ids:
                        self.orig_path = path + '/' + str(max(diff_ids))
                        diff_id = max(diff_ids)+1
                    else:
                        diff_id = 1
                    path = path + '/' + str(diff_id)
                    self.log.info("performing differential backup from '%s' to '%s'", self.orig_path, path)
                    os.makedirs(path)

                # lock backup dir and sync hierarchy
                with open(path+'/lock', 'w') as lockfile:
                    fcntl.flock(lockfile.fileno(), fcntl.LOCK_EX)
                    t0 = time.time()
                    self.backup_hierarchy(path, stats, options, store, user, server, config)
                    changes = stats['changes'] + stats['deletes']
                    self.log.info('backing up %s took %.2f seconds (%d changes, ~%.2f/sec, %d errors)',
                        path, time.time()-t0, changes, changes/(time.time()-t0), stats['errors'])

            # return statistics in output queue
            self.oqueue.put(stats)

    def _store_props(self, store):
        props = list(store.props())

        for proptag in SETTINGS_PROPTAGS:
            if not [prop for prop in props if prop.proptag == proptag]:
                prop = store.get_prop(proptag)
                if prop:
                    props.append(prop)

        return props

    def backup_hierarchy(self, path, stats, options, store, user, server, config):
        # backup user and store properties
        if not options.folders:
            open(path+'/store', 'wb').write(dump_props(self._store_props(store), stats, self.log))
            if user:
                open(path+'/user', 'wb').write(dump_props(user.props(), stats, self.log))
                if not options.skip_meta:
                    open(path+'/delegates', 'wb').write(store.delegations_dumps(stats=stats))
                    open(path+'/acl', 'wb').write(store.permissions_dumps(stats=stats))

        # time of last backup
        open(path+'/timestamp', 'wb').write(pickle_dumps(self.service.timestamp))
        if not os.path.exists(path+'/folders'):
            os.makedirs(path+'/folders')

        # check command-line options and collect folders
        self.log.info('backing up: %s', path)
        sk_folder = {}
        folders = list(store.folders())
        subtree = store.subtree
        if options.recursive:
            folders = sum([[f]+list(f.folders()) for f in folders], [])
        for folder in folders:
            if (not store.public and \
                ((options.skip_junk and folder == store.junk) or \
                 (options.skip_deleted and folder == store.wastebasket))):
                continue
            sk_folder[folder.sourcekey] = folder
        sk_dir = sk_struct(self.orig_path, options)

        # differential
        if options.differential:
            for sk, folder in sk_folder.items():
                data_path = path+'/folders/'+folder.sourcekey
                orig_data_path = (self.orig_path+'/'+sk_dir[sk]) if sk in sk_dir else None
                self.backup_folder(data_path, orig_data_path, folder, subtree, config, options, stats, store, user, server)
            return

        # add new folders
        for new_sk in set(sk_folder) - set(sk_dir):
            data_path = path+'/folders/'+new_sk
            folder = sk_folder[new_sk]
            self.backup_folder(data_path, data_path, folder, subtree, config, options, stats, store, user, server)

        # update existing folders
        for both_sk in set(sk_folder) & set(sk_dir):
            data_path = path+'/'+sk_dir[both_sk]
            folder = sk_folder[both_sk]
            self.backup_folder(data_path, data_path, folder, subtree, config, options, stats, store, user, server)

        # timestamp deleted folders
        if not options.folders:
            for del_sk in set(sk_dir) - set(sk_folder):
                fpath = open(path+'/'+sk_dir[del_sk]+'/path', 'rb').read().decode('utf8')
                index = (path+'/'+sk_dir[del_sk]+'/index')
                _mark_deleted(index, fpath, self.service.timestamp, self.log)

    def backup_folder(self, data_path, orig_data_path, folder, subtree, config, options, stats, store, user, server):
        """ backup single folder """

        self.log.debug('backing up folder: "%s" "%s"', store.name, folder.path)

        if not os.path.isdir(data_path):
            os.makedirs(data_path)

        # backup folder properties, path, metadata
        open(data_path+'/path', 'wb').write(folder.path.encode('utf8'))
        open(data_path+'/folder', 'wb').write(dump_props(folder.props(), stats, self.log))
        if not options.skip_meta:
            open(data_path+'/acl', 'wb').write(folder.permissions_dumps(stats=stats))
            rules = folder.rules_dumps(stats=stats)
            if rules:
                open(data_path+'/rules', 'wb').write(rules)
        if options.only_meta:
            return

        # sync over ICS, using stored 'state'
        importer = FolderImporter(folder, data_path, config, options, self.log, stats, self.service)
        state = None
        if orig_data_path:
            orig_statepath = '%s/state' % orig_data_path
            if os.path.exists(orig_statepath):
                state = open(orig_statepath, 'rb').read()
                self.log.debug('found previous folder sync state: %s', state)
        new_state = folder.sync(importer, state, log=self.log, stats=stats, begin=options.period_begin, end=options.period_end)
        if new_state != state or options.differential:
            importer.commit()
            statepath = '%s/state' % data_path
            open(statepath, 'wb').write(new_state.encode('ascii'))
            self.log.debug('saved folder sync state: %s', new_state)

class FolderImporter:
    """ tracks changes for a given folder """

    def __init__(self, *args):
        self.folder, self.folder_path, self.config, self.options, self.log, self.stats, self.service = args
        self.store = self.folder.store
        self.reset_cache()

    def reset_cache(self):
        self.item_updates = []
        self.index_updates = []
        self.cache_size = 0

    def update(self, item, flags):
        """ store updated item in 'items' database, and subject and date in 'index' database """

        with log_exc(self.log, self.stats):
            self.log.debug('folder %s: new/updated document with entryid %s, sourcekey %s', self.folder.sourcekey, item.entryid, item.sourcekey)

            data = zlib.compress(item.dumps(attachments=not self.options.skip_attachments, archiver=False, skip_broken=True))
            self.item_updates.append((item.sourcekey, data))

            idx = self.make_idx(item, item.read)
            self.index_updates.append((item.sourcekey, idx))

            self.stats['changes'] += 1

            self.cache_size += len(data) + len(idx)
            if self.cache_size > CACHE_SIZE:
                self.commit()

    def make_idx(self, item, read):
        orig_prop = item.get_prop(PR_EC_BACKUP_SOURCE_KEY)
        if orig_prop:
            orig_prop = _hex(orig_prop.value)
        idx = pickle_dumps({
            b'subject': item.subject,
            b'orig_sourcekey': orig_prop,
            b'last_modified': item.last_modified,
            b'backup_updated': self.service.timestamp,
            b'read': read,
        })
        return idx

    def delete(self, item, flags): # XXX batch as well, 'updating' cache?
        """ deleted item from 'items' and 'index' databases """

        with log_exc(self.log, self.stats):
         with closing(dbopen(self.folder_path+'/items')) as db_items:
          with closing(dbopen(self.folder_path+'/index')) as db_index:

            self.log.debug('folder %s: deleted document with sourcekey %s', self.folder.sourcekey, item.sourcekey)

            # NOTE ICS may generate delete events for items that did not exist
            # before, for example for a new message which has already been
            # deleted in the meantime.
            if item.sourcekey.encode('ascii') in db_items:
                idx = pickle_loads(db_index[item.sourcekey.encode('ascii')])
                idx[b'backup_deleted'] = self.service.timestamp
                db_index[item.sourcekey.encode('ascii')] = pickle_dumps(idx)
            else:
                db_index[item.sourcekey.encode('ascii')] = pickle_dumps({
                    b'backup_deleted': self.service.timestamp
                })
            self.stats['deletes'] += 1

    def read(self, item, state):
        idx = self.make_idx(item, state)
        self.index_updates.append((item.sourcekey, idx))

    def commit(self):
        """ commit data to storage """

        t0 = time.time()
        with closing(dbopen(self.folder_path+'/items')) as item_db:
         with closing(dbopen(self.folder_path+'/index')) as index_db:

            for sourcekey, data in self.item_updates:
                item_db[sourcekey.encode('ascii')] = data
            for sourcekey, idx in self.index_updates:
                index_db[sourcekey.encode('ascii')] = idx

        self.log.debug('commit took %.2f seconds (%d items)', time.time()-t0, len(self.item_updates))
        self.reset_cache()

class Service(kopano.Service):
    """ main backup process """

    def main(self):
        self.timestamp = datetime.datetime.now()

        if self.options.restore or (self.options.purge is not None) or self.options.merge:
            data_path = self.args[0].rstrip('/')
            with open(data_path+'/lock', 'w') as lockfile:
                fcntl.flock(lockfile.fileno(), fcntl.LOCK_EX)
                if self.options.restore:
                    self.restore(data_path)
                elif (self.options.purge is not None):
                    self.purge(data_path)
                elif self.options.merge:
                    self.merge(data_path)
        else:
            self.backup()

    def backup(self):
        """ create backup workers, determine which stores to queue, start the workers, display statistics """

        self.iqueue, self.oqueue = Queue(), Queue()
        workers = [BackupWorker(self, 'backup%d'%i, nr=i, iqueue=self.iqueue, oqueue=self.oqueue)
                       for i in range(self.config['worker_processes'])]
        for worker in workers:
            worker.start()
        try:
            jobs = self.create_jobs()
        except kopano.Error as e:
            fatal(str(e))
        for job in jobs:
            self.iqueue.put(job)
        self.log.info('queued %d store(s) for parallel backup (%s processes)', len(jobs), len(workers))
        t0 = time.time()
        stats = [self.oqueue.get() for i in range(len(jobs))] # blocking
        changes = sum(s['changes'] + s['deletes'] for s in stats)
        errors = sum(s['errors'] for s in stats)
        self.log.info('queue processed in %.2f seconds (%d changes, ~%.2f/sec, %d errors)',
            (time.time()-t0), changes, changes/(time.time()-t0), errors)

    def restore(self, data_path):
        """ restore data from backup """

        self.restored_sourcekeys = set()
        self.options.sourcekeys = [sk.upper() for sk in self.options.sourcekeys]

        # determine store to restore to
        self.log.info('starting restore of %s', data_path)
        username = os.path.split(data_path)[1]

        try:
            if self.options.users:
                username = self.options.users[0]
                store = self._store(self.options.users[0])
            elif self.options.stores:
                username = self.options.stores[0]
                store = self.server.store(self.options.stores[0])
            else:
                store = self._store(username)
        except kopano.NotFoundError:
            store = None

        if not store:
            fatal('unable to open store (username: %s)' % username)

        user = store.user

        # determine stored and specified folders
        path_folder = folder_struct(data_path, self.options)
        paths = self.options.folders or sorted(path_folder.keys())
        if self.options.recursive:
            paths = [path2 for path2 in path_folder for path in paths if (path2+'//').startswith(path+'/')]
        for path in paths:
            if path not in path_folder:
                fatal('no such folder: %s' % path)

        # start restore
        self.log.info('restoring to store %s', store.entryid)
        t0 = time.time()
        stats = {'changes': 0, 'errors': 0}

        # determine restore root
        if self.options.restore_root:
            restore_root = store.folder(self.options.restore_root, create=True)
        else:
            restore_root = store.subtree

        # check existing folders
        sk_folder = {}
        for folder in restore_root.folders():
            orig_sk = folder.get(PR_EC_BACKUP_SOURCE_KEY)
            if orig_sk:
                sk_folder[orig_sk] = folder

        # restore specified (parts of) folders
        meta_folders = []
        sks = set()
        for path in paths:
            fpath = path_folder[path]
            folderprops = pickle_loads(open('%s/folder' % fpath, 'rb').read())
            folder_sk = folderprops[PR_SOURCE_KEY]

            # determine folder to restore
            if self.options.sourcekeys:
                with closing(dbopen(fpath+'/items')) as db:
                    if not [sk for sk in self.options.sourcekeys if sk.encode('ascii') in db]:
                        continue
            else:
                if self.options.deletes in (None, 'no') and folder_deleted(fpath):
                    continue

                sks.add(folder_sk)

                folder = restore_root.get_folder(path)
                if (folder and not store.public and \
                    ((self.options.skip_junk and folder == store.junk) or \
                    (self.options.skip_deleted and folder == store.wastebasket))):
                        continue

            # restore folder
            if self.options.only_meta: # TODO create empty folders with acls/rules, or skip non-existing folders?
                folder = restore_root.get_folder(path)
            else:
                # differential folder move
                folder = sk_folder.get(folder_sk)
                if folder and self.options.differential:
                    restore_path = self.options.restore_root+'/'+path if self.options.restore_root else path
                    restore_parent_path = '/'.join(UNESCAPED_SLASH_RE.split(restore_path)[:-1])
                    if folder.parent.path != restore_parent_path:
                        newparent = store.get_folder(restore_parent_path)
                        if newparent:
                            self.log.info('moving folder %s to %s', folder.path, restore_path)
                            folder.parent.move(folder, newparent)
                else:
                    folder = restore_root.folder(path, create=True)

                if self.options.clean_folders:
                    self.log.info('emptying folder %s', folder.path)
                    folder.empty()
                self.restore_folder(folder, path, fpath, store, store.subtree, stats, user, self.server)
            if folder:
                meta_folders.append((folder, fpath))

        # differential folder deletes
        if self.options.differential:
            for sk in set(sk_folder)-sks:
                path = sk_folder[sk].path
                parent = store.get_folder('/'.join(UNESCAPED_SLASH_RE.split(path)[:-1]))
                if parent:
                    self.log.info('deleting folder %s', path)
                    parent.delete(sk_folder[sk])

        # restore folder-level metadata
        if not (self.options.sourcekeys or self.options.skip_meta):
            self.log.info('restoring metadata')
            for (folder, fpath) in meta_folders:
                folder.permissions_loads(open(fpath+'/acl', 'rb').read(), stats=stats)
                if os.path.exists(fpath+'/rules'):
                    folder.rules_loads(open(fpath+'/rules', 'rb').read(), stats=stats)

        # restore store-level metadata (webapp/mapi settings)
        if user and not (self.options.folders or self.options.restore_root or self.options.skip_meta or self.options.sourcekeys):
            if os.path.exists('%s/store' % data_path):
                storeprops = pickle_loads(open('%s/store' % data_path, 'rb').read())
                for proptag in SETTINGS_PROPTAGS:
                    if PROP_TYPE(proptag) == PT_TSTRING:
                        proptag = CHANGE_PROP_TYPE(proptag, PT_UNICODE)
                    value = storeprops.get(proptag)
                    if not value:
                        continue
                    store.mapiobj.SetProps([SPropValue(proptag, value)])
                store.mapiobj.SaveChanges(KEEP_OPEN_READWRITE)
            if os.path.exists('%s/delegates' % data_path):
                store.delegations_loads(open('%s/delegates' % data_path, 'rb').read(), stats=stats)
            if os.path.exists('%s/acl' % data_path):
                store.permissions_loads(open('%s/acl' % data_path, 'rb').read(), stats=stats)

        for sourcekey in self.options.sourcekeys:
            if sourcekey not in self.restored_sourcekeys:
                self.log.error('could not restore sourcekey: %s', sourcekey)

        self.log.info('restore completed in %.2f seconds (%d changes, ~%.2f/sec, %d errors)',
            time.time()-t0, stats['changes'], stats['changes']/(time.time()-t0), stats['errors'])

    def purge(self, data_path):
        """ permanently delete old folders/items from backup """

        stats = {'folders': 0, 'items': 0}
        path_folder = folder_struct(data_path, self.options)

        for path, fpath in path_folder.items():
            # check if folder was deleted
            self.log.info('checking folder: %s', path)
            if folder_deleted(fpath):
                if (self.timestamp - folder_deleted(fpath)).days >= self.options.purge:
                    self.log.debug('purging folder')
                    shutil.rmtree(fpath)
                    stats['folders'] += 1

            else: # check all items for deletion
                items_deleted = False
                with closing(dbopen(fpath+'/items')) as db_items:
                    with closing(dbopen(fpath+'/index')) as db_index:
                        delete_items = []
                        for item, idx in db_index.items():
                            d = pickle_loads(idx)
                            backup_deleted = d.get(b'backup_deleted')
                            if backup_deleted and (self.timestamp - backup_deleted).days >= self.options.purge:
                                delete_items.append(item)

                        if delete_items:
                            items_deleted = True

                        for item in delete_items:
                            stats['items'] += 1
                            self.log.debug('purging item: %s', item)
                            del db_index[item]
                            if item in db_items:
                                del db_items[item]
                        if items_deleted:
                            self.log.debug('compacting index database')
                            freed_pages = db_index.db.compact(flags=bsddb.db.DB_FREE_SPACE)
                            self.log.debug('returned %d pages to the underlying filesystem', freed_pages)
                    if items_deleted:
                        self.log.debug('compacting items database')
                        freed_pages = db_items.db.compact(flags=bsddb.db.DB_FREE_SPACE)
                        self.log.debug('returned %d pages to the underlying filesystem', freed_pages)

        self.log.info('purged %d folders and %d items', stats['folders'], stats['items'])

    def merge(self, data_path):
        """ merge differential backups """

        diff_base = 'differential/'+data_path
        diff_ids = sorted((x for x in os.listdir(diff_base)), key=lambda x: int(x))

        for diff_id in diff_ids:
            diff_path = diff_base+'/'+diff_id
            self.log.info("merging differential backup '%s' into '%s'", diff_path, data_path)

            timestamp = pickle_loads(open(diff_path+'/timestamp', 'rb').read())

            orig_sk_dir = sk_struct(data_path, self.options)
            diff_sk_dir = sk_struct(diff_path, self.options)

            # add new folders
            for new_sk in set(diff_sk_dir) - set(orig_sk_dir):
                from_dir = diff_path+'/'+diff_sk_dir[new_sk]
                to_dir = data_path+'/folders/'+new_sk
                fpath = open(from_dir+'/path', 'rb').read().decode('utf8')
                self.log.debug("merging new folder '%s'", fpath)
                _copy_folder_meta(from_dir, to_dir)

            # update existing folders # XXX check matching & higher syncstate?
            for both_sk in set(orig_sk_dir) & set(diff_sk_dir):
                folder_dir = diff_path+'/'+diff_sk_dir[both_sk]
                fpath = open(folder_dir+'/path', 'rb').read().decode('utf8')
                orig_dir = data_path+'/'+orig_sk_dir[both_sk]

                # now merge new data
                with closing(dbopen(orig_dir+'/items')) as orig_db_items:
                 with closing(dbopen(orig_dir+'/index')) as orig_db_index:
                  with closing(dbopen(folder_dir+'/items')) as diff_db_items:
                   with closing(dbopen(folder_dir+'/index')) as diff_db_index:

                    if diff_db_index: # XXX check higher if file exists
                        self.log.debug("merging '%s' (%d updates)", fpath, len(diff_db_index))

                        for key, value in diff_db_index.items():
                            if key in orig_db_index:
                                idx = pickle_loads(orig_db_index[key])
                                idx.update(pickle_loads(value)) # possibly only update 'backup_deleted' (differential)
                            else:
                                idx = pickle_loads(value)
                            orig_db_index[key] = pickle_dumps(idx)

                            if key in diff_db_items:
                                orig_db_items[key] = diff_db_items[key] # differential may contain pure delete (no item)

                _copy_folder_meta(folder_dir, orig_dir, keep_db=True)

            # timestamp deleted folders
            for del_sk in set(orig_sk_dir) - set(diff_sk_dir):
                orig_dir = data_path+'/'+orig_sk_dir[del_sk]
                fpath = open(orig_dir+'/path', 'rb').read().decode('utf8')

                _mark_deleted(orig_dir+'/index', fpath, timestamp, self.log)

            _copy_store_meta(diff_path, data_path)

    def create_jobs(self):
        """ check command-line options and determine which stores should be backed up """

        output_dir = self.options.output_dir or u''
        jobs = []

        # specified companies/all users
        if self.options.companies or not (self.options.users or self.options.stores):
            servers = self.config['backup_servers']
            if servers is None:
                servers = [self.server.name]
            for company in self.server.companies():
                companyname = company.name if company.name != 'Default' else ''
                for user in company.users():
                    if (not servers or user.home_server in servers) and user.store:
                        jobs.append((user.store, user.name, os.path.join(output_dir, companyname, user.name)))
                public_store = company.public_store
                if public_store and not self.options.skip_public and \
                   (not servers or public_store.home_server in servers):
                    target = 'public@'+companyname if companyname else 'public'
                    jobs.append((company.public_store, None, os.path.join(output_dir, companyname, target)))

        # specified users
        if self.options.users:
            for user in self.server.users():
                if user.store:
                    jobs.append((user.store, user.name, os.path.join(output_dir, user.name)))

        # specified stores
        if self.options.stores:
            for store in self.server.stores():
                if store.public and store.company:
                    if store.company.name != 'Default':
                        target = 'public@' + store.company.name
                    else:
                        target = 'public'
                else:
                    target = store.guid
                jobs.append((store, None, os.path.join(output_dir, target)))

        # check folders
        if self.options.folders:
            for job in jobs:
                list(job[0].folders())

        return [(job[0].entryid,)+job[1:] for job in sorted(jobs, reverse=True, key=lambda x: x[0].size)]

    def restore_folder(self, folder, path, data_path, store, subtree, stats, user, server):
        """ restore (partial) folder """

        folderprops = pickle_loads(open('%s/folder' % data_path, 'rb').read())

        if not self.options.sourcekeys:
            self.log.debug('restoring folder %s', path)

            # restore name, container class
            for proptag in (PR_DISPLAY_NAME_W, PR_CONTAINER_CLASS_W):
                value = folderprops.get(proptag)
                if value is not None:
                    folder[proptag] = value

        # load existing sourcekeys in folder, to check for duplicates
        existing = {}
        table = folder.mapiobj.GetContentsTable(0)
        table.SetColumns([PR_SOURCE_KEY, PR_EC_BACKUP_SOURCE_KEY, PR_ENTRYID], 0)
        for row in table.QueryRows(2147483647, 0):
            if PROP_TYPE(row[1].ulPropTag) != PT_ERROR:
                existing[_hex(row[1].Value)] = _hex(row[2].Value)
            else:
                existing[_hex(row[0].Value)] = _hex(row[2].Value)

        # now dive into 'items', and restore desired items
        with closing(dbopen(data_path+'/items')) as db_items:
          with closing(dbopen(data_path+'/index')) as db_index:
            index = dict((a, pickle_loads(b)) for (a,b) in db_index.items())

            # determine sourcekey(s) to restore
            sourcekeys = db_index.keys()
            if self.options.sourcekeys:
                sourcekeys = [sk for sk in self.options.sourcekeys if sk.encode('ascii') in sourcekeys]
            else:
                sourcekeys = [sk.decode('ascii') for sk in sourcekeys]

            # restore/delete each item
            for sourcekey2 in sourcekeys:
                sourcekey2a = sourcekey2.encode('ascii')
                with log_exc(self.log, stats):

                    # differential delete
                    if index[sourcekey2a].get(b'backup_deleted') and self.options.differential:
                        if self.options.deletes == 'no':
                            self.log.warning('skipping deleted item with sourcekey %s', sourcekey2)
                            continue
                        if sourcekey2a not in existing:
                            self.log.warning('item with sourcekey %s already deleted', sourcekey2)
                            continue

                        # delete item
                        self.log.debug('deleting item with sourcekey %s', sourcekey2)
                        item = folder.item(entryid=existing[sourcekey2a])
                        folder.delete(item)
                        continue

                    # regular delete
                    if(sourcekey2a not in db_items or \
                       (index[sourcekey2a].get(b'backup_deleted') and self.options.deletes in (None, 'no'))
                    ):
                        continue

                    # date range check
                    last_modified = index[sourcekey2a].get(b'last_modified')
                    if(last_modified and \
                       ((self.options.period_begin and last_modified < self.options.period_begin) or \
                         (self.options.period_end and last_modified >= self.options.period_end))
                    ):
                         continue

                    # handle existing item
                    entryid = existing.get(sourcekey2a) or existing.get(index[sourcekey2a][b'orig_sourcekey'])
                    if entryid is not None:
                        if self.options.differential or self.options.overwrite:
                            folder.delete(folder.item(entryid=entryid))
                        else:
                            self.log.warning('skipping duplicate item with sourcekey %s', sourcekey2)
                            continue

                    # restore item
                    self.log.debug('restoring item with sourcekey %s', sourcekey2)
                    data = zlib.decompress(db_items[sourcekey2a])
                    read = index[sourcekey2a].get(b'read')

                    item = folder.create_item(
                        loads=data,
                        attachments=not self.options.skip_attachments,
                        read=read,
                    )

                    # store original item sourcekey or it is lost
                    try:
                        item.prop(PR_EC_BACKUP_SOURCE_KEY)
                    except (MAPIErrorNotFound, kopano.NotFoundError):
                        item.mapiobj.SetProps([SPropValue(PR_EC_BACKUP_SOURCE_KEY, _unhex(sourcekey2))])
                        item.mapiobj.SaveChanges(0)

                    if self.options.sourcekeys:
                        self.restored_sourcekeys.add(sourcekey2)
                    stats['changes'] += 1

        # store original folder sourcekey
        folder_sk = folderprops[PR_SOURCE_KEY]
        try:
            folder.prop(PR_EC_BACKUP_SOURCE_KEY)
        except (MAPIErrorNotFound, kopano.NotFoundError):
            folder.mapiobj.SetProps([SPropValue(PR_EC_BACKUP_SOURCE_KEY, folder_sk)])
            folder.mapiobj.SaveChanges(0)

    def _store(self, username):
        """ lookup store for username """

        if '@' in username:
            u, c = username.split('@')
            if u == 'public' or u == c:
                return self.server.company(c).public_store
            else:
                return self.server.user(username).store
        elif username == 'public':
            return self.server.public_store
        else:
            return self.server.user(username).store

def sk_struct(data_path, options, mapper=None, base_path=None):
    """ determine all folders in backup directory """

    if mapper is None:
        mapper = {}
        base_path = data_path
    if os.path.exists(data_path+'/folders'):
        for f in os.listdir(data_path+'/folders'):
            d = data_path+'/folders/'+f
            if os.path.exists(d+'/path'): # XXX purge empty dirs?
                mapper[d.split('/')[-1]] = d[len(base_path)+1:] # XXX ugly
                sk_struct(d, options, mapper, base_path)
    return mapper

def folder_struct(data_path, options, mapper=None): # XXX deprecate?
    """ determine all folders in backup directory """

    if mapper is None:
        mapper = {}
    if os.path.exists(data_path+'/path'):
        path = open(data_path+'/path', 'rb').read().decode('utf8')
        mapper[path] = data_path
    if os.path.exists(data_path+'/folders'):
        for f in os.listdir(data_path+'/folders'):
            d = data_path+'/folders/'+f
            if os.path.isdir(d):
                folder_struct(d, options, mapper)
    return mapper

def folder_deleted(data_path):
    if os.path.exists(data_path+'/index'):
        with closing(bsddb.hashopen(data_path+'/index')) as db:
           idx = db.get(b'folder')
           if idx and pickle_loads(idx).get(b'backup_deleted'):
               return pickle_loads(idx).get(b'backup_deleted')
    return None

def show_contents(data_path, options):
    """ summary of contents of backup directory, at the item or folder level, in CSV format """

    # setup CSV writer, perform basic checks
    writer = csv.writer(sys.stdout)
    path_folder = folder_struct(data_path, options)
    paths = options.folders or sorted(path_folder)
    for path in paths:
        if path not in path_folder:
            fatal('no such folder: %s' % path)
    if options.recursive:
        paths = [p for p in path_folder if [f for f in paths if p.startswith(f)]]

    # loop over folders
    for path in paths:
        data_path = path_folder[path]
        items = []

        if options.deletes == 'no' and folder_deleted(data_path):
            continue

        # filter items on date using 'index' database
        if os.path.exists(data_path+'/index'):
            with closing(bsddb.hashopen(data_path+'/index')) as db:
                for key, value in db.items():
                    d = pickle_loads(value)
                    if ((key == b'folder') or
                        (options.period_begin and d[b'last_modified'] < options.period_begin) or
                        (options.period_end and d[b'last_modified'] >= options.period_end) or
                        (options.deletes == 'no' and d.get(b'backup_deleted'))):
                        continue
                    if b'last_modified' in d: # ignore sourcekey-only deletes (differential)
                        items.append((key, d))

        # --stats: one entry per folder
        if options.stats:
            writer.writerow([path, len(items)])

        # --index: one entry per item
        elif options.index:
            items = sorted(items, key=lambda item: item[1][b'last_modified'])
            for key, d in items:
                writer.writerow([key.decode('ascii'), path, d[b'last_modified'], d[b'subject']])

def dump_props(props, stats, log):
    """ dump given MAPI properties """

    data = {}
    with log_exc(log, stats):
        data = dict((prop.proptag, prop.mapiobj.Value) for prop in props)
    return pickle_dumps(data)

def main():
    # select common options
    parser = kopano.parser('CKQSufwUPcslObeV', usage='kopano-backup [PATH] [options]')

    # add custom options
    parser.add_option('', '--skip-junk', dest='skip_junk', action='store_true', help='skip junk folder')
    parser.add_option('', '--skip-deleted', dest='skip_deleted', action='store_true', help='skip deleted items folder')
    parser.add_option('', '--skip-public', dest='skip_public', action='store_true', help='skip public store')
    parser.add_option('', '--skip-attachments', dest='skip_attachments', action='store_true', help='skip attachments')
    parser.add_option('', '--skip-meta', dest='skip_meta', action='store_true', help='skip metadata')
    parser.add_option('', '--only-meta', dest='only_meta', action='store_true', help='only backup/restore metadata')
    parser.add_option('', '--deletes', dest='deletes', help='store/restore deleted items/folders', metavar='YESNO')
    parser.add_option('', '--purge', dest='purge', type='int', help='purge items/folders deleted more than N days ago', metavar='N')
    parser.add_option('', '--restore', dest='restore', action='store_true', help='restore from backup')
    parser.add_option('', '--restore-root', dest='restore_root', help='restore under specific folder', metavar='PATH')
    parser.add_option('', '--stats', dest='stats', action='store_true', help='list folders for PATH')
    parser.add_option('', '--index', dest='index', action='store_true', help='list items for PATH')
    parser.add_option('', '--sourcekey', dest='sourcekeys', action='append', default=[], help='restore specific sourcekey', metavar='SOURCEKEY')
    parser.add_option('', '--recursive', dest='recursive', action='store_true', help='backup/restore folders recursively')
    parser.add_option('', '--differential', dest='differential', action='store_true', help='create/restore differential backup')
    parser.add_option('', '--overwrite', dest='overwrite', action='store_true', help='overwrite duplicate items')
    parser.add_option('', '--merge', dest='merge', action='store_true', help='merge differential backups')
    parser.add_option('', '--clean-folders', dest='clean_folders', action='store_true', help='empty folder(s) before restore (dangerous!)')

    # parse and check command-line options
    options, args = parser.parse_args()

    options.service = False
    if options.restore or options.merge or options.stats or options.index or (options.purge is not None):
        if len(args) != 1 or not os.path.isdir(args[0]):
            fatal('please specify path to backup data')
    elif len(args) != 0:
        fatal('too many arguments')
    if options.deletes and options.deletes not in ('yes', 'no'):
        fatal("the --deletes option takes 'yes' or 'no'")
    if options.folders and (options.differential or (options.purge is not None) or options.merge):
        fatal('the --folder option cannot be combined with --differential, --purge or --merge')
    if options.output_dir and options.differential:
        fatal('the --output-dir option cannot be combined with --differential')
    if options.sourcekeys and options.differential:
        fatal('invalid use of --sourcekeys option')

    if options.stats or options.index:
        # handle --stats/--index
        show_contents(args[0], options)
    else:
        # start backup/restore
        Service('backup', options=options, config=CONFIG, args=args).start()

if __name__ == '__main__':
    main() # pragma: no cover



#!/usr/local/bin/python3.14
# SPDX-License-Identifier: AGPL-3.0-or-later

from MAPI.Util import HrGetOneProp, SPropValue
from MAPI.Tags import PR_FINDER_ENTRYID, PR_ENTRYID

import kopano

"""

as of Kopano-Core 8.2, PR_STORE_SUPPORT_MASK has STORE_SEARCH_OK enabled
for every store. this means that this script needs to run at upgrade time,
to create the findroots and ACLS so that cross-store searches will actually
work.

"""

FINDROOT_RIGHTS = ['folder_visible', 'read_items', 'create_subfolders', 'edit_own', 'delete_own']


def upgrade_findroot(store):
    findroot = store.root.get_folder('FINDER_ROOT')
    if not findroot:
        print('creating FINDER_ROOT for store %s' % store.guid)

        # create findroot
        findroot = store.root.folder('FINDER_ROOT', create=True)

        # set PR_FINDER_ENTRYID # XXX pyko: store.findroot = ..
        entryid = HrGetOneProp(findroot.mapiobj, PR_ENTRYID)
        store.mapiobj.SetProps([SPropValue(PR_FINDER_ENTRYID, entryid.Value)])
        store.mapiobj.SaveChanges(0)

    # add findroot ACL
    findroot.permission(kopano.Group('Everyone'), create=True).rights = FINDROOT_RIGHTS

    # we don't need to update PR_STORE_SUPPORT_MASK as it is hardcoded
    # whether it includes STORE_SEARCH_OK..


def main():
    for store in kopano.stores():
        upgrade_findroot(store)
        if store.archive_store:
            upgrade_findroot(store.archive_store)


if __name__ == '__main__':
    main()
